import os
import pandas as pd
import autotrader.indicators as indicators


class TestIndicators:
    def test_supertrend(self):
        testdata = self._get_test_data()
        st = indicators.supertrend(testdata)
        assert st["trend"][-1] == 1, "Supertrend indicator failed"

    def test_halftrend(self):
        testdata = self._get_test_data()
        ht = indicators.halftrend(testdata)
        assert round(ht["halftrend"][-1], 5) == 1.12344, "Halftrend indicator failed"
        assert round(ht["atrHigh"][-1], 5) == 1.12593, "Halftrend indicator failed"
        assert round(ht["atrLow"][-1], 5) == 1.12095, "Halftrend indicator failed"

    def test_range_filter(self):
        testdata = self._get_test_data()
        rfi = indicators.range_filter(testdata)
        assert round(rfi["upper"][-1], 6) == 1.12736, "Range filter indicator failed"
        assert round(rfi["lower"][-1], 6) == 1.119815, "Range filter indicator failed"
        assert round(rfi["rf"][-1], 6) == 1.123588, "Range filter indicator failed"

    @staticmethod
    def test_rolling_signal_list():
        signals = [0, 1, 0, 0, 0, -1, 0, 0, 1, 0, 0]
        rolled_signals = indicators.rolling_signal_list(signals)
        assert rolled_signals == [
            0,
            1,
            1,
            1,
            1,
            -1,
            -1,
            -1,
            1,
            1,
            1,
        ], "Rolling signal list indicator failed"

    @staticmethod
    def test_unroll_signal_list():
        signals = [0, 1, 1, 1, 1, -1, -1, -1, 1, 1, 1]
        unrolled_signals = indicators.unroll_signal_list(signals)
        assert list(unrolled_signals) == [
            0,
            1,
            0,
            0,
            0,
            -1,
            0,
            0,
            1,
            0,
            0,
        ], "Unroll signal list indicator failed"

    @staticmethod
    def _get_test_data():
        home_dir = os.path.abspath(os.path.dirname(__file__))
        data_dir = os.path.join(home_dir, "data", "EUR_USD_H4.csv")
        data = pd.read_csv(data_dir, index_col=0, parse_dates=True)
        testdata = data.iloc[-1000:]
        return testdata



import os
import sys
import time
import pickle
import timeit
import importlib
import traceback
import numpy as np
import pandas as pd
from tqdm import tqdm
from art import tprint
from typing import Callable
from threading import Thread
from ast import literal_eval
from scipy.optimize import brute
from datetime import datetime, timedelta, timezone
from autotrader.autoplot import AutoPlot
from autotrader.autobot import AutoTraderBot
from autotrader.utilities import (
    read_yaml,
    get_broker_config,
    DataStream,
    TradeAnalysis,
    unpickle_broker,
)


class AutoTrader:
    """A Python-Based Development Platform For Automated Trading Systems.

    Methods
    -------
    configure(...)
        Configures run settings for AutoTrader.
    add_strategy(...)
        Adds a strategy to the active AutoTrader instance.
    backtest(...)
        Configures backtest settings.
    optimise(...)
        Configures optimisation settings.
    scan(...)
        Configures scan settings.
    run()
        Runs AutoTrader with configured settings.
    add_data(...)
        Specify local data files to use for backtests.
    plot_settings(...)
        Configures the plot settings for AutoPlot.
    get_bots_deployed(instrument=None)
        Returns the AutoTrader trading bots deployed in the active instance.
    plot_backtest(bot=None)
        Plots backtest results of a trading Bot.
    plot_multibot_backtest(trade_results=None)
        Plots backtest results for multiple trading bots.
    multibot_backtest_analysis(bots=None)
        Analyses backtest results of multiple trading bots.
    print_trade_results(trade_results)
        Prints trade results.
    print_multibot_trade_results(trade_results=None)
        Prints a multi-bot backtest results.

    References
    ----------
    Author: Kieran Mackle

    Homepage: https://kieran-mackle.github.io/AutoTrader/

    GitHub: https://github.com/kieran-mackle/AutoTrader
    """

    def __init__(self) -> None:
        """AutoTrader initialisation. Called when creating new AutoTrader
        instance.
        """
        # Public attributes
        self.trade_results = None

        self._home_dir = None
        self._verbosity = 1

        self._global_config_dict = None
        self._instance_str = None
        self._run_mode = "continuous"
        self._papertrading = False
        self._timestep = None
        self._strategy_timestep = None
        self._warmup_period = pd.Timedelta("0s").to_pytimedelta()
        self._feed = None
        self._req_liveprice = False
        self._max_workers = None

        # Communications
        self._notify = 0
        self._email_params = None
        self._order_summary_fp = None

        # Livetrade Parameters
        self._deploy_time = None
        self._check_data_alignment = True
        self._allow_dancing_bears = False
        self._maintain_broker_thread = False

        # Broker parameters
        self._execution_method = None  # Execution method
        self._broker = None  # Broker instance(s)
        self._brokers_dict = None  # Dictionary of brokers
        self._broker_name = ""  # Broker name(s)
        self._broker_utils = None  # Broker utilities
        self._broker_verbosity = 0  # Broker verbosity
        self._environment = "paper"  # Trading environment
        self._account_id = None  # Trading account
        self._base_currency = None
        self._no_brokers = 0
        self._multiple_brokers = False

        # Strategy parameters
        self._strategy_configs = {}
        self._strategy_classes = {}
        self._shutdown_methods = {}
        self._uninitiated_strat_files = []
        self._uninitiated_strat_dicts = []
        self._bots_deployed = []

        # Backtesting Parameters
        self._backtest_mode = False
        self._data_start = None
        self._data_end = None
        self._broker_histories = None

        # Local Data Parameters
        self._data_directory = None
        self._data_indexing = "open"
        self._data_stream_object = DataStream
        self._data_file = None
        self._MTF_data_files = None
        self._data_path_mapper = None
        self._local_data = None
        self._local_quote_data = None
        self._auxdata = None
        self._dynamic_data = False
        self._allow_duplicate_bars = False

        # Virtual Broker Parameters
        self._virtual_broker_config = {}
        self._virtual_tradeable_instruments = {}  # Instruments tradeable mapper
        self._broker_refresh_freq = "1s"

        # Optimisation Parameters
        self._optimise_mode = False
        self._opt_params = None
        self._bounds = None
        self._Ns = None

        # Scan Parameters
        self._scan_mode = False
        self._scan_index = None
        self._scan_watchlist = None

        # Plotting
        self._show_plot = False
        self._max_indis_over = 3
        self._max_indis_below = 2
        self._fig_tools = "pan,wheel_zoom,box_zoom,undo,redo,reset,save,crosshair"
        self._ohlc_height = 400
        self._ohlc_width = 800
        self._top_fig_height = 150
        self._bottom_fig_height = 150
        self._jupyter_notebook = False
        self._show_cancelled = True
        self._chart_timeframe = "default"
        self._chart_theme = "caliber"
        self._use_strat_plot_data = False
        self._plot_portolio_chart = False

    def __repr__(self):
        return f"AutoTrader instance"

    def __str__(self):
        return "AutoTrader instance"

    def configure(
        self,
        verbosity: int = 1,
        broker: str = None,
        execution_method: Callable = None,
        feed: str = None,
        req_liveprice: bool = False,
        notify: int = 0,
        home_dir: str = None,
        allow_dancing_bears: bool = False,
        account_id: str = None,
        environment: str = "paper",
        show_plot: bool = False,
        jupyter_notebook: bool = False,
        mode: str = "continuous",
        update_interval: str = None,
        data_index_time: str = "open",
        global_config: dict = None,
        instance_str: str = None,
        broker_verbosity: int = 0,
        home_currency: str = None,
        allow_duplicate_bars: bool = False,
        deploy_time: datetime = None,
        max_workers: int = None,
    ) -> None:
        """Configures run settings for AutoTrader.

        Parameters
        ----------
        verbosity : int, optional
            The verbosity of AutoTrader (0, 1, 2). The default is 1.
        broker : str, optional
            The broker(s) to connect to for trade execution. Multiple exchanges
            can be provided using comma separattion. The default is 'virtual'.
        execution_method : Callable, optional
            The execution model to call when submitting orders to the broker.
            This method must accept the broker instance, the order object,
            order_time and any *args, **kwargs.
        feed : str, optional
            The data feed to be used. This can be the same as the broker
            being used, or another data source. Options include 'yahoo',
            'oanda', 'ib', 'dydx', 'ccxt', 'local' or 'none'. When data is provided
            via the add_data method, the feed is automatically set to 'local'.
            The default is None.
        req_liveprice : bool, optional
            Request live market price from broker before placing trade, rather
            than using the data already provided. The default is False.
        notify : int, optional
            The level of email notifications (0, 1, 2). The default is 0.
        home_dir : str, optional
            The project home directory. The default is the current working directory.
        allow_dancing_bears : bool, optional
            Allow incomplete candles to be passed to the strategy. The default is False.
        account_id : str, optional
            The brokerage account ID to be used. The default is None.
        environment : str, optional
            The trading environment of this instance ('paper', 'live'). The
            default is 'paper'.
        show_plot : bool, optional
            Automatically generate trade chart. The default is False.
        jupyter_notebook : bool, optional
            Set to True when running in Jupyter notebook environment. The
            default is False.
        mode : str, optional
            The run mode (either 'periodic' or 'continuous'). The default is
            'periodic'.
        update_interval : str, optional
            The update interval to use when running in 'continuous' mode. This
            should align with the highest resolution bar granularity in your
            strategy to allow adequate updates. The string inputted will be
            converted to a timedelta object. If None is passed, the update
            interval will be inferred from the strategy INTERVAL. The
            default is None.
        data_index_time : str, optional
            The time by which the data is indexed. Either 'open', if the data
            is indexed by the bar open time, or 'close', if the data is indexed
            by the bar close time. The default is 'open'.
        global_config : dict, optional
            Optionally provide your global configuration directly as a
            dictionary, rather than it being read in from a yaml file. The
            default is None.
        instance_str : str, optional
            The name of the active AutoTrader instance, used to control bots
            deployed when livetrading in continuous mode. When not specified,
            the instance string will be of the form 'autotrader_instance_n'.
            The default is None.
        broker_verbosity : int, optional
            The verbosity of the broker. The default is 0.
        home_currency : str, optional
            The home currency of trading accounts used (intended for FX
            conversions). The default is None.
        allow_duplicate_bars : bool, optional
            Allow duplicate bars to be passed on to the strategy. The default
            is False.
        deploy_time : datetime, optional
            The time to deploy the bots. If this is a future time, AutoTrader
            will wait until it is reached before deploying. It will also be used
            as an anchor to synchronise future bot updates. If not specified,
            bots will be deployed as soon as possible, with successive updates
            synchronised to the deployment time.
        max_workers : int, optional
            The maximum number of workers to use when spawning threads. The
            default is None.

        Returns
        -------
        None
            Calling this method configures the internal settings of
            the active AutoTrader instance.
        """
        self._verbosity = verbosity
        self._feed = feed
        self._req_liveprice = req_liveprice
        self._broker_name = broker if broker is not None else self._broker_name
        self._execution_method = execution_method
        self._notify = notify
        self._home_dir = home_dir if home_dir is not None else os.getcwd()
        self._allow_dancing_bears = allow_dancing_bears
        self._allow_duplicate_bars = allow_duplicate_bars
        self._account_id = account_id
        self._environment = environment
        self._show_plot = show_plot
        self._jupyter_notebook = jupyter_notebook
        self._run_mode = mode
        self._timestep = (
            pd.Timedelta(update_interval).to_pytimedelta()
            if update_interval is not None
            else self._timestep
        )
        self._data_indexing = data_index_time
        self._global_config_dict = global_config
        self._instance_str = instance_str
        self._broker_verbosity = broker_verbosity
        self._base_currency = home_currency
        self._deploy_time = deploy_time
        self._max_workers = max_workers

    def add_strategy(
        self,
        config_filename: str = None,
        config_dict: dict = None,
        strategy=None,
        shutdown_method: str = None,
    ) -> None:
        """Adds a strategy to AutoTrader.

        Parameters
        ----------
        config_filename : str, optional
            The prefix of the yaml strategy configuration file, located in
            home_dir/config. The default is None.
        config_dict : dict, optional
            Alternative to config_filename, a strategy configuration
            dictionary can be passed directly. The default is None.
        strategy : AutoTrader Strategy, optional
            The strategy class object. The default is None.
        shutdown_method : str, optional
            The name of the shutdown method in the strategy (if any). This
            method will be called when AutoTrader is livetrading in continuous
            mode, and the instance has recieved the shutdown signal. The
            default is None.

        Returns
        -------
        None
            The strategy will be added to the active AutoTrader instance.
        """
        # TODO - assign unique ID to different strategies to prevent instrument
        # traded names conflict

        if self._home_dir is None:
            # Home directory has not yet been set, postpone strategy addition
            if config_filename is None:
                self._uninitiated_strat_dicts.append(config_dict)
            else:
                self._uninitiated_strat_files.append(config_filename)

            if shutdown_method is not None:
                raise Exception(
                    "Providing the shutdown method requires "
                    + "the home directory to have been configured. "
                    + "please either specify it, or simply call "
                    + "the configure method before adding a strategy."
                )

        else:
            # Home directory has been set, continue
            if config_dict is None:
                # Config YAML filepath provided
                config_file_path = os.path.join(
                    self._home_dir, "config", config_filename
                )
                new_strategy = read_yaml(config_file_path + ".yaml")
            else:
                # Config dictionary provided directly
                new_strategy = config_dict

            # Construct strategy name
            try:
                name = new_strategy["NAME"]
            except KeyError:
                raise Exception(
                    "Please specify the name of your strategy via the "
                    + "'NAME' key of the strategy configuration. This name should "
                    + "match the class name of your strategy."
                )

            # Check for other required keys
            # TODO - review required keys
            required_keys = ["CLASS", "INTERVAL", "PERIOD"]
            for key in required_keys:
                if key not in new_strategy:
                    print(
                        f"Please include the '{key}' key in your strategy configuration."
                    )
                    sys.exit(0)

            if name in self._strategy_configs:
                print(
                    "Warning: duplicate strategy name deteced. Please check "
                    + "the NAME field of your strategy configuration file and "
                    + "make sure it is not the same as other strategies being "
                    + "run from this instance."
                )
                print("Conflicting name:", name)

            # Check PERIOD type
            if isinstance(new_strategy["PERIOD"], float):
                # Convert to integer
                new_strategy["PERIOD"] = int(new_strategy["PERIOD"])

            elif isinstance(new_strategy["PERIOD"], str):
                # Period provided as time range, convert to int using INTERVAL
                new_strategy["PERIOD"] = int(
                    pd.Timedelta(new_strategy["PERIOD"])
                    / pd.Timedelta(new_strategy["INTERVAL"])
                )

            # Save to AutoTrader instance
            self._strategy_configs[name] = new_strategy
            self._shutdown_methods[name] = shutdown_method

            # Set timestep from strategy config
            try:
                strat_granularity = pd.Timedelta(
                    new_strategy["INTERVAL"]
                ).to_pytimedelta()
            except:
                print(
                    f"Strategy configuration error: invalid time interval: '{new_strategy['INTERVAL']}'."
                )
                sys.exit(0)

            if self._strategy_timestep is None:
                # Timestep hasn't been set yet; set it
                self._strategy_timestep = strat_granularity

            else:
                # Timestep has been set, overwrite only with a smaller granularity
                if strat_granularity < self._strategy_timestep:
                    self._strategy_timestep = strat_granularity

        if strategy is not None:
            self._strategy_classes[strategy.__name__] = strategy

    def virtual_account_config(
        self,
        verbosity: int = 0,
        initial_balance: float = 1000,
        spread: float = 0,
        commission: float = 0,
        spread_units: str = "price",
        commission_scheme: str = "percentage",
        maker_commission: float = None,
        taker_commission: float = None,
        leverage: int = 1,
        hedging: bool = False,
        margin_call_fraction: float = 0,
        default_slippage_model: Callable = None,
        slippage_models: dict = None,
        picklefile: str = None,
        exchange: str = None,
        tradeable_instruments: list = None,
        refresh_freq: str = "1s",
        home_currency: str = None,
        papertrade: bool = True,
    ) -> None:
        """Configures the virtual broker's initial state to allow livetrading
        on the virtual broker. If you wish to create multiple virtual broker
        instances, call this method for each virtual account.

        Parameters
        ----------
        verbosity : int, optional
            The verbosity of the broker. The default is 0.
        initial_balance : float, optional
            The initial balance of the account. The default is 1000.
        spread : float, optional
            The bid/ask spread to use in backtest (specified in units defined
            by the spread_units argument). The default is 0.
        spread_units : str, optional
            The unit of the spread specified. Options are 'price', meaning that
            the spread is quoted in price units, or 'percentage', meaning that
            the spread is quoted as a percentage of the market price. The default
            is 'price'.
        commission : float, optional
            Trading commission as percentage per trade. The default is 0.
        commission_scheme : str, optional
            The method with which to apply commissions to trades made. The options
            are (1) 'percentage', where the percentage specified by the commission
            argument is applied to the notional trade value, (2) 'fixed_per_unit',
            where the monetary value specified by the commission argument is
            multiplied by the number of units in the trade, and (3) 'flat', where
            a flat monetary value specified by the commission argument is charged
            per trade made, regardless of size. The default is 'percentage'.
        maker_commission : float, optional
            The commission to charge on liquidity-making orders. The default is
            None, in which case the nominal commission argument will be used.
        taker_commission: float, optional
            The commission to charge on liquidity-taking orders. The default is
            None, in which case the nominal commission argument will be used.
        leverage : int, optional
            Account leverage. The default is 1.
        hedging : bool, optional
            Allow hedging in the virtual broker (opening simultaneous
            trades in oposing directions). The default is False.
        margin_call_fraction : float, optional
            The fraction of margin usage at which a margin call will occur.
            The default is 0.
        default_slippage_model : Callable, optional
            The default model to use when calculating the percentage slippage
            on the fill price, for a given order size. The default functon
            returns zero.
        slippage_models : dict, optional
            A dictionary of callable slippage models, keyed by instrument.
        picklefile : str, optional
            The filename of the picklefile to load state from. If you do not
            wish to load from state, leave this as None. The default is None.
        exchange : str, optional
            The name of the exchange to use for execution. This gets passed to
            an instance of AutoData to update prices and use the realtime
            orderbook for virtual order execution. The default is None.
        tradeable_instruments : list, optional
            A list containing strings of the instruments tradeable through the
            exchange specified. This is used to determine which exchange orders
            should be submitted to when trading across multiple exchanges. This
            should account for all instruments provided in the watchlist. The
            default is None.
        refresh_freq : str, optional
            The timeperiod to sleep for in between updates of the virtual broker
            data feed when manually papertrading. The default is '1s'.
        home_currency : str, optional
            The home currency of the account. The default is None.
        papertrade : bool, optional
            A boolean to flag when the account is to be used for papertrading
            (real-time trading on paper). The default is True.
        """

        # TODO - allow specifying spread dictionary to have custom spreads for
        # different products

        # Enforce virtual broker and paper trading environment
        if exchange is not None:
            broker_name = exchange
        else:
            # TODO - catch attempt to create multiple instances with same exchange
            broker_name = "virtual"

        if broker_name != "virtual" and broker_name not in self._broker_name:
            # Unrecognised broker
            print(
                f"Please specify the broker '{broker_name}' in the "
                + "configure method before configuring its virtual account."
            )
            sys.exit(0)
            # self._broker_name += broker_name + ', '

        self._papertrading = False if self._backtest_mode else papertrade
        self._broker_refresh_freq = refresh_freq

        if tradeable_instruments is not None:
            self._virtual_tradeable_instruments[broker_name] = tradeable_instruments

        # Construct configuration dictionary
        config = {
            # Broker configuration
            "verbosity": verbosity,
            "initial_balance": initial_balance,
            "leverage": leverage,
            "spread": spread,
            "spread_units": spread_units,
            "commission": commission,
            "commission_scheme": commission_scheme,
            "maker_commission": maker_commission,
            "taker_commission": taker_commission,
            "hedging": hedging,
            "base_currency": home_currency,
            "paper_mode": self._papertrading,
            "public_trade_access": False,  # Not yet implemented
            "margin_closeout": margin_call_fraction,
            "default_slippage_model": default_slippage_model,
            "slippage_models": slippage_models,
            "picklefile": picklefile,
            # Extra parameters
            "execution_feed": exchange,
        }

        # Append
        # TODO - check keys for already existing
        self._virtual_broker_config[broker_name] = config

    def backtest(
        self,
        start: str = None,
        end: str = None,
        start_dt: datetime = None,
        end_dt: datetime = None,
        warmup_period: str = "0s",
    ) -> None:
        """Configures settings for backtesting.

        Parameters
        ----------
        start : str, optional
            Start date for backtesting, in format dd/mm/yyyy. The default is None.
        end : str, optional
            End date for backtesting, in format dd/mm/yyyy. The default is None.
        start_dt : datetime, optional
            Datetime object corresponding to start time. The default is None.
        end_dt : datetime, optional
            Datetime object corresponding to end time. The default is None.
        warmup_period : str, optional
            A string describing the warmup period to be used. This is
            equivalent to the minimum period of time required to collect
            sufficient data for the strategy. The default is '0s'.

        Notes
        ------
            Start and end times must be specified as the same type. For
            example, both start and end arguments must be provided together,
            or alternatively, start_dt and end_dt must both be provided.
        """
        # Convert start and end strings to datetime objects
        if start_dt is None and end_dt is None:
            start_dt = datetime.strptime(start + "+0000", "%d/%m/%Y%z")
            end_dt = datetime.strptime(end + "+0000", "%d/%m/%Y%z")

        # Assign attributes
        self._backtest_mode = True
        self._data_start = start_dt
        self._data_end = end_dt
        self._warmup_period = pd.Timedelta(warmup_period).to_pytimedelta()

    def optimise(
        self, opt_params: list, bounds: list, Ns: int = 4, force_download: bool = False
    ) -> None:
        """Optimisation configuration.

        Parameters
        ----------
        opt_params : list
            The parameters to be optimised, as they  are named in the
            strategy configuration file.
        bounds : list(tuples)
            The bounds on each of the parameters to be optimised, specified
            as a tuple of the form (lower, upper) for each parameter. The
            default is 4.
        force_download : bool, optional
            Force AutoTrader to download data each iteration. This is not
            recommended. Instead, you should provide local download to optimise
            on, using the add_data method. The default is False.
        Ns : int, optional
            The number of points along each dimension of the optimisation grid.

        Raises
        ------
        Exception:
            When force_download is False, and local data has not been added
            through the add_data method. Note that add_data should be called
            prior to calling the optimise method.

        Returns
        -------
        None:
            The optimisation settings will be saved to the active AutoTrader
            instance.

        See Also
        --------
        AutoTrader.add_data()
        """

        if type(bounds) == str:
            full_tuple = literal_eval(bounds)
            bounds = [(x[0], x[-1]) for x in full_tuple]

        if type(opt_params) == str:
            opt_params = opt_params.split(",")

        self._optimise_mode = True
        self._opt_params = opt_params
        self._bounds = bounds
        self._Ns = Ns

        if self._local_data is None:
            raise Exception(
                "Local data files have not been provided. "
                + "Please do so using AutoTrader.add_data(), "
                + "or set force_download to True to proceed."
            )

    def add_data(
        self,
        data_dict: dict = None,
        mapper_func: callable = None,
        quote_data: dict = None,
        data_directory: str = "price_data",
        abs_dir_path: str = None,
        auxdata: dict = None,
        stream_object=None,
        dynamic_data: bool = False,
    ) -> None:
        """Specify local data to run AutoTrader on.

        Parameters
        ----------
        data_dict : dict, optional
            A dictionary containing the filenames of the datasets
            to be used. The default is None.
        mapper_func : callable, optional
            A callable used to provide the absolute filepath to the data
            given the instrument name (as it appears in the watchlist)
            as an input argument. The default is None.
        quote_data : dict, optional
            A dictionary containing the quote data filenames
            of the datasets provided in data_dict. The default is None.
        data_directory : str, optional
            The name of the sub-directory containing price
            data files. This directory should be located in the project
            home directory (at.home_dir). The default is 'price_data'.
        abs_dir_path : str, optional
            The absolute path to the data_directory. This parameter
            may be used when the datafiles are stored outside of the project
            directory. The default is None.
        auxdata : dict, optional
            A dictionary containing the data paths to supplement the
            data passed to the strategy module. For strategies involving
            multiple products, the keys of this dictionary must correspond
            to the products, with the auxdata in nested dictionaries or
            otherwise. The default is None.
        stream_object : DataStream, optional
            A custom data stream object, allowing custom data pipelines. The
            default is DataStream (from autotrader.utilities).
        dynamic_data : bool, optional
            A boolean flag to signal that the stream object provided should
            be refreshed each timestep of a backtest. This can be useful when
            backtesting strategies with futures contracts, which expire and
            must be rolled. The default is False.

        Raises
        ------
        Exception
            When multiple quote-data files are provided per instrument traded.

        Returns
        -------
        None
            Data will be assigned to the active AutoTrader instance for
            later use.

        Notes
        ------
            To ensure proper directory configuration, this method should only
            be called after calling autotrader.configure().

            The data provided to the strategy will either contain a single
            timeframe OHLC dataframe, a dictionary of MTF dataframes, or
            a dict with 'base' and 'aux' keys, for aux and base strategy
            data (which could be single of MTF).

        Examples
        --------
            An example data_dict is shown below.

            >>> data_dict = {'product1': 'filename1.csv',
                             'product2': 'filename2.csv'}

            For MTF data, data_dict should take the form shown below. In
            the case of MTF data, quote data should only be provided for
            the base timeframe (ie. the data which will be iterated on
            when backtesting). Therefore, the quote_data dict will look
            the same for single timeframe and MTF backtests.

            >>> data_dict = {'product1': {'H1': 'product1_H1.csv',
                                          'D': 'product1_D.csv'},
                             'product2': {'H1': 'product2_H1.csv',
                                          'D': 'product2_D.csv'}
                             }

            An example for the quate_data dictionary is shown below.

            >>> quote_data = {'product1': 'product1_quote.csv',
                              'product2': 'product2_quote.csv'}

            The auxdata dictionary can take the form shown below. This data
            will be passed on to your strategy.

            >>> auxdata = {'product1': 'aux_price_data.csv',
                           'product2': {'extra_data1': 'dataset1.csv',
                                        'extra_data2': 'dataset2.csv'}
                           }

        """
        self._data_directory = data_directory
        dir_path = (
            abs_dir_path
            if abs_dir_path is not None
            else os.path.join(self._home_dir, data_directory)
        )

        # Trading data
        if data_dict is not None:
            # Assign local data attribute
            local_data = {}

            # Populate local_data
            for product in data_dict:
                if type(data_dict[product]) == dict:
                    # MTF data
                    MTF_data = {}
                    for timeframe in data_dict[product]:
                        MTF_data[timeframe] = os.path.join(
                            dir_path, data_dict[product][timeframe]
                        )

                    local_data[product] = MTF_data
                else:
                    # Single timeframe data
                    local_data[product] = os.path.join(dir_path, data_dict[product])

            self._local_data = local_data

        # Quote data
        if quote_data is not None:
            # Assign local data attribute
            local_quote_data = {}

            # Populate local_quote_data
            for product in quote_data:
                if type(quote_data[product]) == dict:
                    raise Exception(
                        "Only a single quote-data file should be "
                        + "provided per instrument traded."
                    )

                else:
                    local_quote_data[product] = os.path.join(
                        dir_path, quote_data[product]
                    )

            self._local_quote_data = local_quote_data

        if auxdata is not None:
            modified_auxdata = {}
            for product, item in auxdata.items():
                if isinstance(item, dict):
                    # Multiple datasets for product
                    modified_auxdata[product] = {}
                    for key, dataset in item:
                        modified_auxdata[product][key] = os.path.join(dir_path, dataset)
                else:
                    # Item is not dict
                    modified_auxdata[product] = os.path.join(dir_path, item)

            self._auxdata = modified_auxdata

        if mapper_func is not None:
            self._data_path_mapper = mapper_func

        # Assign data stream object
        if stream_object is not None:
            self._data_stream_object = stream_object

        # Fix attributes
        self._feed = "local"
        self._dynamic_data = dynamic_data

    def scan(
        self,
        strategy_filename: str = None,
        strategy_dict: dict = None,
        scan_index: str = None,
    ) -> None:
        """Configure AutoTrader scan settings.

        Parameters
        ----------
        strategy_filename : str, optional
             The prefix of yaml strategy configuration file, located in
             home_dir/config. The default is None.
        strategy_dict : dict, optional
            A strategy configuration dictionary. The default is None.
        scan_index : str, optional
            Forex scan index. The default is None.

        Returns
        -------
        None
            The scan settings of the active AutoTrader instance will be
            configured.
        """

        # If a strategy is provided here, add it
        if strategy_filename is not None:
            self.add_strategy(strategy_filename)
        elif strategy_dict is not None:
            self.add_strategy(strategy_dict=strategy_dict)

        # If scan index provided, use that. Else, use strategy watchlist
        scan_index = "Strategy watchlist"

        self._scan_mode = True
        self._scan_index = scan_index
        self._check_data_alignment = False

    def run(self) -> None:
        """Performs essential checks and runs AutoTrader."""
        # Print Banner
        if int(self._verbosity) > 0:
            tprint("AutoTrader", font="tarty1")

        # Define home_dir if undefined
        if self._home_dir is None:
            self._home_dir = os.getcwd()

        # Load uninitiated strategies
        for strat_dict in self._uninitiated_strat_dicts:
            self.add_strategy(strategy_dict=strat_dict)
        for strat_config_file in self._uninitiated_strat_files:
            self.add_strategy(strategy_filename=strat_config_file)

        if self._scan_watchlist is not None:
            # Scan watchlist has not overwritten strategy watchlist
            self._update_strategy_watchlist()

        # Check for added strategies
        if (
            len(self._strategy_configs) == 0
            and not self._papertrading
            and int(self._verbosity) > 1
        ):
            print(
                "Warning: no strategy has been provided. Do so by using the"
                + " 'add_strategy' method of AutoTrader."
            )

        if sum([self._backtest_mode, self._scan_mode]) > 1:
            print(
                "Error: backtest mode and scan mode are both set to True,"
                + " but only one of these can run at a time."
            )
            print("Please check your inputs and try again.")
            sys.exit(0)

        # Check self._timestep
        if self._timestep is None:
            # Set timestep based on strategy inferred timestep
            self._timestep = self._strategy_timestep

        # Remove any trailing commas in self._broker_name
        self._broker_name = self._broker_name.strip().strip(",")

        # Check for multiple brokers
        self._no_brokers = len(self._broker_name.split(","))
        self._multiple_brokers = self._no_brokers > 1
        # TODO - check len(self._virtual_broker_config)

        # Check self._broker_name
        if self._broker_name == "":
            # Broker has not been assigned
            if self._backtest_mode or self._papertrading:
                # Use virtual broker
                self._broker_name = "virtual"

            else:
                # Livetrading
                raise Exception(
                    "Please specify the name(s) of the broker(s) "
                    + "you wish to trade with."
                )

        if self._backtest_mode:
            if self._notify > 0:
                print(
                    "Warning: notify set to {} ".format(self._notify)
                    + "during backtest. Setting to zero to prevent emails."
                )
                self._notify = 0

            # Check that the backtest does not request future data
            if self._data_end > datetime.now(tz=self._data_end.tzinfo):
                print(
                    "Warning: you have requested backtest data into the "
                    + "future. The backtest end date will be adjsuted to "
                    + "the current time."
                )
                self._data_end = datetime.now(tz=self._data_end.tzinfo)

            # Check if the broker has been configured
            if len(self._virtual_broker_config) != self._no_brokers:
                # Virtual accounts have not been configured for the brokers specified
                if len(self._virtual_broker_config) == 0:
                    # Use default values for all virtual accounts
                    for exchange in self._broker_name.split(","):
                        self.virtual_account_config(papertrade=False, exchange=exchange)
                else:
                    # Partially configured accounts
                    raise Exception(
                        "Please configure the virtual accounts for "
                        + "each broker you plan to used.\n"
                        + f" Number of brokers specifed: {self._no_brokers}\n"
                        + f" Number of virtual accounts configured: {len(self._virtual_broker_config)}"
                    )

        # Preliminary checks complete, continue initialisation
        if self._optimise_mode:
            # Run optimisation
            if self._backtest_mode:
                self._run_optimise()
            else:
                print("Please set backtest parameters to run optimisation.")

        else:
            # Trading
            if not self._backtest_mode and "virtual" in self._broker_name:
                # Not in backtest mode, yet virtual broker is selected
                if not self._papertrading:
                    # Not papertrading either
                    raise Exception(
                        "Live-trade mode requires setting the "
                        + "broker. Please do so using the "
                        + "AutoTrader configure method. If you "
                        + "would like to use the virtual broker "
                        + "for papertrading, please "
                        + "configure the virtual broker account(s) "
                        + "with the virtual_account_config method."
                    )

            # Load global (account) configuration
            if self._global_config_dict is not None:
                # Use global config dict provided
                global_config = self._global_config_dict
            else:
                # Try load from file
                global_config_fp = os.path.join(self._home_dir, "config", "keys.yaml")
                if os.path.isfile(global_config_fp):
                    global_config = read_yaml(global_config_fp)
                else:
                    global_config = None

                # Assign
                self._global_config_dict = global_config

            # Check data feed requirements
            if self._feed is None:
                # No data feed specified
                if self._backtest_mode:
                    raise Exception(
                        "No data feed specified! Please do so using "
                        + "AutoTrader.configure(feed=), or provide local data via "
                        + "AutoTrader.add_data()."
                    )

            elif global_config is None and self._feed.lower() in ["oanda", "ib"]:
                # No global configuration provided, but data feed requires authentication
                raise Exception(
                    f'Data feed "{self._feed}" requires global '
                    + "configuration. If a config file already "
                    + "exists, make sure to specify the home_dir. "
                    + "Alternatively, provide a configuration dictionary "
                    + "directly via AutoTrader.configure()."
                )

            # Check global config requirements
            if sum([self._backtest_mode, self._scan_mode, self._papertrading]) == 0:
                # Livetrade mode
                if global_config is None:
                    # No global_config
                    raise Exception(
                        "No global configuration found (required for "
                        + "livetrading). Either provide a global configuration dictionary "
                        + "via the configure method, or create a keys.yaml file in your "
                        + "config/ directory."
                    )

                if self._broker_name == "":
                    raise Exception(
                        "Please specify the brokers you would like to "
                        + "trade with via the configure method."
                    )

                # Check broker
                supported_exchanges = ["virtual", "oanda", "ib", "ccxt", "dydx"]
                inputted_brokers = self._broker_name.lower().replace(" ", "").split(",")
                for broker in inputted_brokers:
                    if broker.split(":")[0] not in supported_exchanges:
                        raise Exception(
                            f"Unsupported broker requested: {self._broker_name}\n"
                            + "Please check the broker(s) specified in configure method and "
                            + "virtual_account_config."
                        )

            # All checks passed, proceed to run main
            if self._verbosity > 1:
                print("All preliminary checks complete, proceeding.")
            self._main()

            if self._papertrading or len(self._bots_deployed) == 0:
                # Return broker instance
                self.broker = self._broker
                return self._broker

    def plot_settings(
        self,
        max_indis_over: int = 3,
        max_indis_below: int = 2,
        fig_tools: str = "pan,wheel_zoom,box_zoom,undo,redo,reset,save,crosshair",
        ohlc_height: int = 400,
        ohlc_width: int = 800,
        top_fig_height: int = 150,
        bottom_fig_height: int = 150,
        jupyter_notebook: bool = False,
        show_cancelled: bool = True,
        chart_timeframe: str = "default",
        chart_theme: str = "caliber",
        use_strat_plot_data: bool = False,
        portfolio_chart: bool = False,
    ) -> None:
        """Configure the plot settings.

        Parameters
        ----------
        max_indis_over : int, optional
            Maximum number of indicators overlaid on the main chart. The
            default is 3.
        max_indis_below : int, optional
            Maximum number of indicators below the main chart. The default is 2.
        fig_tools : str, optional
            The figure tools. The default is "pan,wheel_zoom,box_zoom,undo,
            redo,reset,save,crosshair".
        ohlc_height : int, optional
            The height (px) of the main chart. The default is 400.
        ohlc_width : int, optional
            The width (px) of the main chart. The default is 800.
        top_fig_height : int, optional
            The height (px) of the figure above the main chart. The default is 150.
        bottom_fig_height : int, optional
            The height (px) of the figure(s) below the main chart. The default is 150.
        jupyter_notebook : bool, optional
            Boolean flag when running in Jupyter Notebooks, to allow inline
            plotting. The default is False.
        show_cancelled : bool, optional
            Show/hide cancelled trades. The default is True.
        chart_timeframe : str, optional
            The bar timeframe to use when gerating the chart. The timeframe
            provided must be a part of the strategy dataset. The default is 'default'.
        chart_theme : bool, optional
            The theme of the Bokeh chart generated. The default is "caliber".
        use_strat_plot_data : bool, optional
            Boolean flag to use data from the strategy instead of candlestick
            data for the chart. If True, ensure your strategy has a timeseries
            data attribute named 'plot_data'. The default is False.
        portfolio_chart : bool, optional
            Override the default plot settings to plot the portfolio chart
            even when running a single instrument backtest.

        Returns
        -------
        None
            The plot settings will be saved to the active AutoTrader instance.
        """
        # Assign attributes
        self._max_indis_over = max_indis_over
        self._max_indis_below = max_indis_below
        self._fig_tools = fig_tools
        self._ohlc_height = ohlc_height
        self._ohlc_width = ohlc_width
        self._top_fig_heigh = top_fig_height
        self._bottom_fig_height = bottom_fig_height
        self._jupyter_notebook = jupyter_notebook
        self._show_cancelled = show_cancelled
        self._chart_timefram = chart_timeframe
        self._chart_theme = chart_theme
        self._use_strat_plot_data = use_strat_plot_data
        self._plot_portolio_chart = portfolio_chart

    def get_bots_deployed(self, instrument: str = None) -> dict:
        """Returns a dictionary of AutoTrader trading bots, organised by
        instrument traded.

        Parameters
        ----------
        instrument : str, optional
            The instrument of the bot to retrieve. The default is None.

        Returns
        -------
        dict
            A dictionary of deployed AutoTrader bot instances.

        Notes
        -----
        If there is only one trading bot deployed, this will be returned
        directly, rather than in a dict.
        """
        bots = {}
        for bot in self._bots_deployed:
            symbol = bot.instrument
            if isinstance(symbol, list):
                # Porfolio-trading bot
                symbol = "portfolio"
            bots[symbol] = bot

        if instrument is not None:
            # Retrieve bot requested
            try:
                bots = bots[instrument]
            except:
                raise Exception(
                    f"There were no bots found to be trading '{instrument}'."
                )
        else:
            if len(bots) == 1:
                # Single bot backtest, return directly
                bots = bots[list(bots.keys())[0]]

        return bots

    def print_trade_results(self, trade_results: TradeAnalysis = None) -> None:
        """Prints trade results.

        Parameters
        ----------
        trade_results : TradeAnalysis
            The trade analysis results class object.

        Returns
        -------
        None
            Trade results will be printed.
        """
        if trade_results is None:
            trade_results = self.trade_results

        trade_summary = trade_results.summary()
        if "start" in trade_summary:
            start_date = trade_summary["start"].strftime("%b %d %Y %H:%M:%S")
            end_date = trade_summary["end"].strftime("%b %d %Y %H:%M:%S")
            duration = trade_summary["end"] - trade_summary["start"]

            starting_balance = trade_summary["starting_balance"]
            ending_balance = trade_summary["ending_balance"]
            ending_NAV = trade_summary["ending_NAV"]
            abs_return = trade_summary["abs_return"]
            pc_return = trade_summary["pc_return"]
            max_drawdown = trade_summary["max_drawdown"]

            no_trades = trade_summary["no_trades"]
            no_long_trades = trade_summary["no_long_trades"]
            no_short_trades = trade_summary["no_short_trades"]
            if no_trades > 0:
                win_rate = trade_summary["all_trades"]["win_rate"]
                max_win = trade_summary["all_trades"]["max_win"]
                avg_win = trade_summary["all_trades"]["avg_win"]
                max_loss = trade_summary["all_trades"]["max_loss"]
                avg_loss = trade_summary["all_trades"]["avg_loss"]
                longest_win_streak = trade_summary["all_trades"]["win_streak"]
                longest_lose_streak = trade_summary["all_trades"]["lose_streak"]
                total_fees = trade_summary["all_trades"]["total_fees"]
                total_volume = trade_summary["all_trades"]["total_volume"]
                adv = total_volume / duration.days

            print("\n----------------------------------------------")
            print("               Trading Results")
            print("----------------------------------------------")
            print("Start date:              {}".format(start_date))
            print("End date:                {}".format(end_date))
            print("Duration:                {}".format(duration))
            print("Starting balance:        ${}".format(round(starting_balance, 2)))
            print("Ending balance:          ${}".format(round(ending_balance, 2)))
            print("Ending NAV:              ${}".format(round(ending_NAV, 2)))
            print(
                "Total return:            ${} ({}%)".format(
                    round(abs_return, 2), round(pc_return, 1)
                )
            )
            print("Maximum drawdown:        {}%".format(round(max_drawdown * 100, 2)))
            if no_trades > 0:
                print("Total no. trades:        {}".format(no_trades))
                print("No. long trades:         {}".format(no_long_trades))
                print("No. short trades:        {}".format(no_short_trades))
                print("Total fees paid:         ${}".format(round(total_fees, 3)))
                print("Total volume traded:     ${}".format(round(total_volume, 2)))
                print("Average daily volume:    ${}".format(round(adv, 2)))
                # print("Win rate:                {}%".format(round(win_rate, 1)))
                # print("Max win:                 ${}".format(round(max_win, 2)))
                # print("Average win:             ${}".format(round(avg_win, 2)))
                # print("Max loss:                -${}".format(round(max_loss, 2)))
                # print("Average loss:            -${}".format(round(avg_loss, 2)))
                # print(
                #     "Longest winning streak:  {} positions".format(longest_win_streak)
                # )
                # print(
                #     "Longest losing streak:   {} positions".format(longest_lose_streak)
                # )
                # if len(trade_results.position_summary) > 0:
                #     avg_pos_dur = np.mean(
                #         trade_results.position_summary.loc["avg_duration"].values
                #     )
                #     print(
                #         "Avg. position duration:  {}".format(
                #             avg_pos_dur,
                #         )
                #     )

                no_open = trade_summary["no_open"]
                if no_open > 0:
                    print("Positions still open:    {}".format(no_open))

            else:
                print("\n No trades made.")

            no_cancelled = trade_summary["no_cancelled"]
            if no_cancelled > 0:
                print("Cancelled orders:        {}".format(no_cancelled))

            # if len(trade_results.position_summary) > 0:
            #     # Long positions
            #     no_long = trade_results.position_summary.sum(axis=1, numeric_only=True)[
            #         "no_long"
            #     ]
            #     print("\n          Summary of long positions")
            #     print("----------------------------------------------")
            #     if no_long > 0:
            #         # avg_long_win = trade_summary["long_positions"]["avg_long_win"]
            #         # max_long_win = trade_summary["long_positions"]["max_long_win"]
            #         # avg_long_loss = trade_summary["long_positions"]["avg_long_loss"]
            #         # max_long_loss = trade_summary["long_positions"]["max_long_loss"]
            #         # long_wr = trade_summary["long_positions"]["long_wr"]
            #         avg_long_dur = np.mean(
            #             trade_results.position_summary.loc["avg_long_duration"].values
            #         )

            #         print("No. long positions:      {}".format(no_long))
            #         print("Avg. position duration:  {}".format(avg_long_dur))
            #         # print("Win rate:                {}%".format(round(long_wr, 1)))
            #         # print("Max win:                 ${}".format(round(max_long_win, 2)))
            #         # print("Average win:             ${}".format(round(avg_long_win, 2)))
            #         # print("Max loss:                -${}".format(round(max_long_loss, 2)))
            #         # print("Average loss:            -${}".format(round(avg_long_loss, 2)))
            #     else:
            #         print("There were no long positions.")

            #     # Short trades
            #     no_short = trade_results.position_summary.sum(
            #         axis=1, numeric_only=True
            #     )["no_short"]
            #     print("\n         Summary of short positions")
            #     print("----------------------------------------------")
            #     if no_short > 0:
            #         # avg_short_win = trade_summary["short_positions"]["avg_short_win"]
            #         # max_short_win = trade_summary["short_positions"]["max_short_win"]
            #         # avg_short_loss = trade_summary["short_positions"]["avg_short_loss"]
            #         # max_short_loss = trade_summary["short_positions"]["max_short_loss"]
            #         # short_wr = trade_summary["short_positions"]["short_wr"]
            #         avg_short_dur = np.mean(
            #             trade_results.position_summary.loc["avg_long_duration"].values
            #         )

            #         print("No. short positions:     {}".format(no_short))
            #         print("Avg. position duration:  {}".format(avg_short_dur))
            #         # print("Win rate:                {}%".format(round(short_wr, 1)))
            #         # print("Max win:                 ${}".format(round(max_short_win, 2)))
            #         # print("Average win:             ${}".format(round(avg_short_win, 2)))
            #         # print("Max loss:                -${}".format(round(max_short_loss, 2)))
            #         # print("Average loss:            -${}".format(round(avg_short_loss, 2)))

            #     else:
            #         print("There were no short positions.")

            # Check for multiple instruments
            if len(trade_results.instruments_traded) > 1:
                # Mutliple instruments traded
                instruments = trade_results.instruments_traded
            #     trade_history = trade_results.isolated_position_history

            #     total_no_trades = []
            #     max_wins = []
            #     max_losses = []
            #     avg_wins = []
            #     avg_losses = []
            #     profitable_trades = []
            #     win_rates = []
            #     for i in range(len(instruments)):
            #         instrument_trades = trade_history[
            #             trade_history.instrument == instruments[i]
            #         ]
            #         no_trades = len(instrument_trades)
            #         total_no_trades.append(no_trades)
            #         max_wins.append(instrument_trades.profit.max())
            #         max_losses.append(instrument_trades.profit.min())
            #         avg_wins.append(
            #             instrument_trades.profit[instrument_trades.profit > 0].mean()
            #         )
            #         avg_losses.append(
            #             instrument_trades.profit[instrument_trades.profit < 0].mean()
            #         )
            #         profitable_trades.append((instrument_trades.profit > 0).sum())
            #         win_rates.append(
            #             100 * profitable_trades[i] / no_trades if no_trades > 0 else 0.0
            #         )

            #     results = pd.DataFrame(
            #         data={
            #             "Instrument": instruments,
            #             "Max. Win": max_wins,
            #             "Max. Loss": max_losses,
            #             "Avg. Win": avg_wins,
            #             "Avg. Loss": avg_losses,
            #             "Win Rate": win_rates,
            #         }
            #     ).fillna(0)

            #     print("\n Instrument Breakdown:")
            #     print(results.to_string(index=False))

        else:
            print("No updates to report.")

    def plot_backtest(self, bot=None) -> None:
        """Plots trade results of an AutoTrader Bot.

        Parameters
        ----------
        bot : AutoTrader bot instance, optional
            AutoTrader bot class containing trade results. The default
            is None.

        Returns
        -------
        None
            A chart will be generated and shown.
        """

        def portfolio_plot():
            ap = self._instantiate_autoplot()
            ap._portfolio_plot(self.trade_results)

        def single_instrument_plot(bot):
            data = bot._check_strategy_for_plot_data(self._use_strat_plot_data)
            ap = self._instantiate_autoplot(data)
            ap.plot(trade_results=bot.trade_results)

        if bot is None:
            # No bot has been provided, select automatically
            if (
                len(self.trade_results.instruments_traded) > 1
                or len(self._bots_deployed) > 1
                or self._plot_portolio_chart
            ):
                # Multi-bot backtest
                portfolio_plot()
            else:
                # Single bot backtest
                bot = self._bots_deployed[0]
                single_instrument_plot(bot)
        else:
            # A bot has been provided
            single_instrument_plot(bot)

    def _main(self) -> None:
        """Run AutoTrader with configured settings."""
        # Check tradeable instruments
        if (
            self._multiple_brokers
            and len(self._virtual_tradeable_instruments) != self._no_brokers
            and self._backtest_mode
        ):
            raise Exception(
                "Please define the tradeable instruments for "
                + "each virtual account configured."
            )

        # Get broker configuration
        if self._backtest_mode or self._papertrading:
            names_list = [f"virtual:{i}" for i in self._broker_name.split(",")]
            broker_names = ",".join(names_list)
        else:
            broker_names = self._broker_name
        broker_config = get_broker_config(
            global_config=self._global_config_dict,
            broker=broker_names,
            environment=self._environment,
        )

        if self._account_id is not None:
            if self._multiple_brokers:
                raise Exception(
                    "Cannot use provided account ID when "
                    + "trading across multiple exchanges. Please specify the "
                    + "desired account in the keys config."
                )
            else:
                # Overwrite default account in config dicts
                broker_config["ACCOUNT_ID"] = self._account_id
                self._global_config_dict["custom_account_id"] = self._account_id

        # Append broker verbosity to broker_config
        if self._multiple_brokers:
            for broker, config in broker_config.items():
                config["verbosity"] = self._broker_verbosity
        else:
            broker_config["verbosity"] = self._broker_verbosity

        # Connect to exchanges
        if self._verbosity > 1:
            print("Connecting to exchanges...")
        self._assign_broker(broker_config)
        if self._verbosity > 1:
            print("  Done.")

        # Configure emailing
        self._configure_emailing(self._global_config_dict)

        # Initialise broker histories
        self._broker_histories = {
            key: {
                "NAV": [],
                "equity": [],
                "margin": [],
                "open_interest": [],
                "long_exposure": [],
                "short_exposure": [],
                "long_unrealised_pnl": [],
                "short_unrealised_pnl": [],
                "long_pnl": [],
                "short_pnl": [],
                "time": [],
            }
            for key in self._brokers_dict
        }

        # Assign trading bots to each strategy
        if self._verbosity > 1:
            print("Spawning trading bots...")
        for strategy, config in self._strategy_configs.items():
            # Check for portfolio strategy
            portfolio = config["PORTFOLIO"] if "PORTFOLIO" in config else False
            watchlist = [config["WATCHLIST"]] if portfolio else config["WATCHLIST"]
            for instrument in watchlist:
                if portfolio:
                    data_dict = self._local_data
                    quote_data_path = self._local_quote_data
                    auxdata = self._auxdata
                else:
                    data_dict = (
                        self._local_data[instrument]
                        if self._local_data is not None
                        else None
                    )
                    quote_data_path = (
                        self._local_quote_data[instrument]
                        if self._local_quote_data is not None
                        else None
                    )
                    auxdata = (
                        self._auxdata[instrument] if self._auxdata is not None else None
                    )

                strategy_class = config["CLASS"]
                strategy_dict = {
                    "config": config,
                    "class": self._strategy_classes[strategy_class]
                    if strategy_class in self._strategy_classes
                    else None,
                    "shutdown_method": self._shutdown_methods[strategy],
                }
                bot = AutoTraderBot(
                    instrument,
                    strategy_dict,
                    self._broker,
                    self._data_start,
                    data_dict,
                    quote_data_path,
                    auxdata,
                    self,
                )
                self._bots_deployed.append(bot)

        if int(self._verbosity) > 0:
            if self._backtest_mode:
                print("BACKTEST MODE")
            else:
                if self._scan_mode:
                    print("SCAN MODE")
                elif self._papertrading:
                    trade_mode = "auto" if len(self._bots_deployed) > 0 else "manual"
                    extra_str = f"{trade_mode} trade in {self._environment} environment"
                    print(f"PAPERTRADE MODE ({extra_str})")
                else:
                    trade_mode = "auto" if len(self._bots_deployed) > 0 else "manual"
                    extra_str = f"{trade_mode} trade in {self._environment} environment"
                    print(f"LIVETRADE MODE ({extra_str})")
                print(
                    "Current time: {}".format(
                        datetime.now().strftime("%A, %B %d %Y, " + "%H:%M:%S")
                    )
                )

            for bot in self._bots_deployed:
                if isinstance(bot.instrument, str):
                    instr_str = bot.instrument
                else:
                    instr_str = (
                        bot.instrument
                        if len(bot.instrument) < 5
                        else f"a portfolio of {len(bot.instrument)} instruments"
                    )
                print(
                    f"\nAutoTraderBot assigned to trade {instr_str}",
                    f"with {bot._broker_name} broker using {bot._strategy_name}.",
                )

        # Begin trading
        self._trade_update_loop()

    def _clear_strategies(self) -> None:
        """Removes all strategies saved in autotrader instance."""
        self._strategy_configs = {}

    def _clear_bots(self) -> None:
        """Removes all deployed bots in autotrader instance."""
        self._bots_deployed = []

    def _instantiate_autoplot(self, data: pd.DataFrame = None) -> AutoPlot:
        """Creates instance of AutoPlot.

        Parameters
        ----------
        data : pd.DataFrame
            The data to instantiate AutoPlot with.

        Raises
        ------
        Exception
            When attempting to plot on missing data timeframe.

        Returns
        -------
        AutoPlot
            An instance of AutoPlot.
        """
        if self._chart_timeframe == "default":
            ap = AutoPlot(data)
        else:
            # Instantiate AutoPlot with requested chart timeframe
            if self._chart_timeframe in self._bots_deployed[0].MTF_data.keys():
                # Valid timeframe requested
                ap = AutoPlot(self._bots_deployed[0].MTF_data[self._chart_timeframe])
                ap._add_backtest_price_data(
                    data
                )  # provide nominal timeframe data for merge operations
            else:
                warning_str = (
                    f"The chart timeframe requested ({self._chart_timeframe}) was not found "
                    + "in the MTF data. Please ensure that the timeframe provided matches "
                    + "the format provided in the strategy configuration file, or the local "
                    + "data provided."
                )
                raise Exception(warning_str)

        # Assign attributes
        ap.configure(
            max_indis_over=self._max_indis_over,
            max_indis_below=self._max_indis_below,
            fig_tools=self._fig_tools,
            ohlc_height=self._ohlc_height,
            ohlc_width=self._ohlc_width,
            top_fig_height=self._top_fig_height,
            bottom_fig_height=self._bottom_fig_height,
            jupyter_notebook=self._jupyter_notebook,
            show_cancelled=self._show_cancelled,
            chart_theme=self._chart_theme,
            use_strat_plot_data=self._use_strat_plot_data,
        )

        return ap

    def _update_strategy_watchlist(self) -> None:
        """Updates the watchlist of each strategy with the scan watchlist."""
        for strategy in self._strategy_configs:
            self._strategy_configs[strategy]["WATCHLIST"] = self._scan_watchlist

    def _assign_broker(self, broker_config: dict) -> None:
        """Configures and assigns appropriate broker(s) for trading."""
        # Check for multiple brokers
        if not self._multiple_brokers:
            # Put broker config in dict to allow single iteration
            broker_config = {self._broker_name: broker_config}

        # Instantiate brokers
        brokers = {}
        brokers_utils = {}
        for broker_key, config in broker_config.items():
            # Import relevant broker and utilities modules
            if self._backtest_mode or self._papertrading:
                # Use virtual broker
                broker_name = "virtual"
            else:
                # Use real broker
                broker_name = broker_key.lower().split(":")[0]

            # Construct utils args
            utils_args = {}
            utils_name = broker_key.lower().split(":")[0]
            if utils_name == "ccxt":
                utils_args["exchange"] = broker_key.lower().split(":")[1]

            # Import relevant modules
            broker_module = importlib.import_module(
                f"autotrader.brokers.{broker_name}.broker"
            )
            utils_module = importlib.import_module(
                f"autotrader.brokers.{utils_name}.utils"
            )

            # Create broker and utils instances
            utils = utils_module.Utils(**utils_args)
            broker = broker_module.Broker(config, utils)

            if self._backtest_mode or self._papertrading:
                # Using virtual broker, configure account
                try:
                    account_config = self._virtual_broker_config[broker_key]
                except KeyError:
                    # Broker hasn't been configured properly
                    raise Exception(
                        f"Broker '{broker_key}' has not been "
                        + "configured. Please do so using the virtual_account_config "
                        + f"method, making sure to specify exchange='{broker_key}'."
                    )

                execution_feed = account_config["execution_feed"]
                feed = self._feed if execution_feed is None else execution_feed
                autodata_config = {
                    "feed": feed,
                    "environment": self._environment,
                    "global_config": self._global_config_dict,
                    "allow_dancing_bears": self._allow_dancing_bears,
                    "base_currency": self._base_currency,
                }
                broker.configure(**account_config, autodata_config=autodata_config)

            # Append to brokers dict
            brokers[broker_key] = broker
            brokers_utils[broker_key] = utils

        # Save broker dict
        self._brokers_dict = brokers

        # Check
        if not self._multiple_brokers:
            # Extract single broker
            brokers = broker
            brokers_utils = utils

        self._broker = brokers
        self._broker_utils = brokers_utils

    def _configure_emailing(self, global_config: dict) -> None:
        """Configure email settings."""
        if int(self._notify) > 0:
            host_email = None
            mailing_list = None

            if "EMAILING" in global_config:
                # Look for host email and mailing list in strategy config, if it
                # was not picked up in strategy config
                if "MAILING_LIST" in global_config["EMAILING"] and mailing_list is None:
                    mailing_list = global_config["EMAILING"]["MAILING_LIST"]
                if "HOST_ACCOUNT" in global_config["EMAILING"] and host_email is None:
                    host_email = global_config["EMAILING"]["HOST_ACCOUNT"]

            if host_email is None:
                print("Warning: email host account not provided.")
            if mailing_list is None:
                print("Warning: no mailing list provided.")

            email_params = {"mailing_list": mailing_list, "host_email": host_email}
            self._email_params = email_params

            logfiles_path = os.path.join(self._home_dir, "logfiles")
            order_summary_fp = os.path.join(logfiles_path, "order_history.txt")

            if not os.path.isdir(logfiles_path):
                os.mkdir(logfiles_path)

            self._order_summary_fp = order_summary_fp

    def _run_optimise(self) -> None:
        """Runs optimisation of strategy parameters."""

        # Modify verbosity for optimisation
        verbosity = self._verbosity
        self._verbosity = 0
        self._show_plot = False
        self.objective = "profit + MDD"

        # Look in self._strategy_configs for config
        if len(self._strategy_configs) > 1:
            print("Error: please optimise one strategy at a time.")
            print("Exiting.")
            sys.exit(0)
        else:
            config_dict = self._strategy_configs[list(self._strategy_configs.keys())[0]]

        my_args = (config_dict, self._opt_params, self._verbosity)

        start = timeit.default_timer()
        result = brute(
            func=self._optimisation_helper_function,
            ranges=self._bounds,
            args=my_args,
            Ns=self._Ns,
            full_output=True,
        )
        stop = timeit.default_timer()

        opt_params = result[0]
        opt_value = result[1]

        print("\nOptimisation complete.")
        print("Time to run: {}s".format(round((stop - start), 3)))
        print("Optimal parameters:")
        print(opt_params)
        print("Objective:")
        print(opt_value)

        # Reset verbosity
        self._verbosity = verbosity

    def _optimisation_helper_function(
        self, params: list, config_dict: dict, opt_params: list, verbosity: int
    ) -> float:
        """Helper function for optimising strategy parameters in AutoTrader.
        This function will parse the ordered params into the config dict.
        """
        for parameter in config_dict["PARAMETERS"]:
            if parameter in opt_params:
                config_dict["PARAMETERS"][parameter] = params[
                    opt_params.index(parameter)
                ]
            else:
                continue

        self._clear_strategies()
        self._clear_bots()
        self.add_strategy(config_dict=config_dict)
        self._main()

        try:
            trade_results = self.trade_results.summary()
            objective = -trade_results["all_trades"]["ending_NAV"]
        except:
            objective = 1000

        print("Parameters/objective:", params, "/", round(objective, 3))

        return objective

    def _check_bot_data(self) -> None:
        """Function to compare lengths of bot data."""
        data_lengths = [len(bot.data) for bot in self._bots_deployed]
        if min(data_lengths) != np.mean(data_lengths):
            print(
                "Warning: mismatched data lengths detected. "
                + "Correcting via row reduction."
            )
            self._normalise_bot_data()
            print("  Done.\n")

    def _normalise_bot_data(self) -> None:
        """Function to normalise the data of mutliple bots so that their
        indexes are equal, allowing backtesting in periodic update mode.
        """

        # Construct list of bot data
        data = [bot.data for bot in self._bots_deployed]

        for i, dat in enumerate(data):
            # Initialise common index
            comm_index = dat.index

            # Update common index by intersection with other data
            for j, dat_2 in enumerate(data):
                comm_index = comm_index.intersection(dat_2.index)

            # Adjust bot data using common indexes
            adj_data = dat[dat.index.isin(comm_index)]

            # Re-assign bot data
            self._bots_deployed[i]._replace_data(adj_data)

    def _get_instance_id(self, dir_name: str = "active_bots"):
        """Returns an ID for the active AutoTrader instance."""
        dirpath = os.path.join(self._home_dir, dir_name)

        # Check if active_bots directory exists
        if not os.path.isdir(dirpath):
            # Directory does not exist, create it
            os.mkdir(dirpath)
            instance_id = 1

        else:
            # Directory exists, find highest instance
            instances = [
                f
                for f in os.listdir(dirpath)
                if os.path.isfile(os.path.join(dirpath, f))
            ]

            last_id = 0
            for instance in instances:
                if "autotrader_instance_" in instance:
                    # Ignore custom instance strings
                    last_id = int(instance.split("_")[-1])

            instance_id = last_id + 1

        return instance_id

    def _check_instance_file(
        self,
        instance_str: str,
        initialisation: bool = False,
        dir_name: str = "active_bots",
        live_check: bool = True,
    ):
        """Checks if the AutoTrader instance exists."""
        if initialisation:
            # Create the file
            filepath = os.path.join(self._home_dir, dir_name, instance_str)
            with open(filepath, mode="w") as f:
                f.write("This instance of AutoTrader contains the following bots:\n")
                for bot in self._bots_deployed:
                    f.write(bot._strategy_name + f" ({bot.instrument})\n")
            instance_file_exists = True

            if int(self._verbosity) > 0 and live_check:
                print(f"Active AutoTrader instance file: active_bots/{instance_str}")

        else:
            dirpath = os.path.join(self._home_dir, dir_name)
            instances = [
                f
                for f in os.listdir(dirpath)
                if os.path.isfile(os.path.join(dir_name, f))
            ]
            instance_file_exists = instance_str in instances

        if int(self._verbosity) > 0 and not instance_file_exists and live_check:
            print(
                f"Instance file '{instance_str}' deleted. AutoTrader",
                "will now shut down.",
            )

        return instance_file_exists

    def _manualtrade(self):
        """Runs the broker updates when manual trading."""
        # Check trading environment
        if self._papertrading:
            # Toggle broker monitoring on
            print(
                f"Running virtual broker updates at {self._broker_refresh_freq} intervals."
            )
            print("To stop papertrading, use at.shutdown().")

            self._maintain_broker_thread = True
            sleep_time = pd.Timedelta(self._broker_refresh_freq).total_seconds()

            # Check for multiple brokers
            if not self._multiple_brokers:
                brokers = {self._broker_name: self._broker}
            else:
                brokers = self._broker

            # Run update loop
            while self._maintain_broker_thread:
                try:
                    for broker_name, broker in brokers.items():
                        broker._update_all()
                        time.sleep(sleep_time)
                except Exception as e:
                    print(e)
            else:
                if int(self._verbosity) > 0:
                    print("Broker update thread killed.")

    def shutdown(self):
        """Shutdown the active AutoTrader instance."""

        if int(self._verbosity) > 0 and self._backtest_mode:
            backtest_end_time = timeit.default_timer()

        # Kill broker update thread
        self._maintain_broker_thread = False

        # Run strategy-specific shutdown routines
        for bot in self._bots_deployed:
            bot._strategy_shutdown()

        # Run instance shut-down routine
        if self._backtest_mode:
            # Create overall backtest results
            if len(self._bots_deployed) == 1:
                price_history = self._bots_deployed[0].data
            else:
                price_history = None
            self.trade_results = TradeAnalysis(
                broker=self._broker,
                broker_histories=self._broker_histories,
                price_history=price_history,
            )

            # Create trade results for each bot
            for bot in self._bots_deployed:
                bot._create_trade_results(broker_histories=self._broker_histories)

            if int(self._verbosity) > 0:
                print(
                    "Backtest complete (runtime "
                    + f"{round((backtest_end_time - self._backtest_start_time), 3)} s)."
                )
                self.print_trade_results()

            if self._show_plot and len(self.trade_results.trade_history) > 0:
                self.plot_backtest()

        elif self._scan_mode and self._show_plot:
            # Show plots for scanned instruments
            for bot in self._bots_deployed:
                ap = self._instantiate_autoplot(bot.data)
                ap.plot(indicators=bot.strategy.indicators, instrument=bot.instrument)
                time.sleep(0.3)

        else:
            # Live trade complete, run livetrade specific shutdown routines
            if self._broker_name.lower() == "ib":
                self._broker._disconnect()

            elif self._papertrading:
                # Paper trade through virtual broker
                self.trade_results = TradeAnalysis(
                    broker=self._broker, broker_histories=self._broker_histories
                )
                self.print_trade_results(self.trade_results)

                picklefile_list = [
                    config["picklefile"] if config["picklefile"] is not None else ""
                    for _, config in self._virtual_broker_config.items()
                ]
                picklefiles = "\n ".join(picklefile_list)
                check_str = picklefiles.strip().split("\n")
                if len(check_str) > 1 or check_str[0] != "":
                    print(
                        f"\nThe following pickle files have been created:\n {picklefiles}"
                        + "\nUse the `unpickle_broker` utility to access these."
                    )

                # Plotting
                if self._show_plot:
                    ap = self._instantiate_autoplot()
                    ap._portfolio_plot(self.trade_results)

    def _trade_update_loop(self):
        """Runs the mode-dependent trade update loop."""
        if int(self._verbosity) > 0 and self._backtest_mode:
            print("\nTrading...\n")
            self._backtest_start_time = timeit.default_timer()

        if len(self._bots_deployed) == 0:
            # No strategy was added; manual trading
            broker_thread = Thread(target=self._manualtrade)
            broker_thread.start()

        else:
            # Automated trading
            if self._run_mode.lower() == "continuous":
                # Running in continuous update mode
                if self._backtest_mode:
                    # Backtesting
                    end_time = self._data_end  # datetime
                    timestamp = self._data_start + self._warmup_period  # datetime
                    pbar = tqdm(
                        total=int((self._data_end - timestamp).total_seconds()),
                        position=0,
                        leave=True,
                    )
                    while timestamp <= end_time:
                        # Update each bot with latest data to generate signal
                        for bot in self._bots_deployed:
                            bot._update(timestamp=timestamp)

                        # Update histories
                        for name, broker in self._brokers_dict.items():
                            hist_dict = self._broker_histories[name]
                            hist_dict["NAV"].append(broker._NAV)
                            hist_dict["equity"].append(broker._equity)
                            hist_dict["margin"].append(broker._margin_available)
                            hist_dict["open_interest"].append(broker._open_interest)
                            hist_dict["long_exposure"].append(broker._long_exposure)
                            hist_dict["short_exposure"].append(broker._short_exposure)
                            hist_dict["long_unrealised_pnl"].append(
                                broker._long_unrealised_pnl
                            )
                            hist_dict["short_unrealised_pnl"].append(
                                broker._short_unrealised_pnl
                            )
                            hist_dict["long_pnl"].append(broker._long_realised_pnl)
                            hist_dict["short_pnl"].append(broker._short_realised_pnl)
                            hist_dict["time"].append(timestamp)

                        # Iterate through time
                        timestamp += self._timestep
                        pbar.update(self._timestep.total_seconds())
                    pbar.close()

                else:
                    # Live trading
                    instance_id = self._get_instance_id()
                    instance_str = (
                        f"autotrader_instance_{instance_id}"
                        if self._instance_str is None
                        else self._instance_str
                    )
                    instance_file_exists = self._check_instance_file(instance_str, True)

                    # Get deploy timestamp
                    if self._deploy_time is not None:
                        deploy_time = self._deploy_time.timestamp()
                        if self._verbosity > 0 and datetime.now() < self._deploy_time:
                            print(f"\nDeploying bots at {self._deploy_time}.")

                    else:
                        deploy_time = time.time()

                    # Wait until deployment time
                    while datetime.now().timestamp() < deploy_time - 0.5:
                        time.sleep(0.5)

                    while instance_file_exists:
                        # Bot instance file exists
                        if self._verbosity > 0:
                            print(f"\nUpdating trading bots.")

                        try:
                            # Update bots
                            # TODO - threadpool executor
                            for bot in self._bots_deployed:
                                try:
                                    # TODO - why UTC? Allow setting manually
                                    bot._update(timestamp=datetime.now(timezone.utc))

                                    if int(self._verbosity) > 0:
                                        print(
                                            f"\nBot update complete: {bot._strategy_name}"
                                        )

                                except:
                                    if int(self._verbosity) > 0:
                                        print(
                                            "Error: failed to update bot running "
                                            + f"{bot._strategy_name} ({bot.instrument})"
                                        )
                                        traceback.print_exc()

                            if self._papertrading:
                                # Update broker histories
                                for name, broker in self._brokers_dict.items():
                                    hist_dict = self._broker_histories[name]
                                    hist_dict["NAV"].append(broker._NAV)
                                    hist_dict["equity"].append(broker._equity)
                                    hist_dict["margin"].append(broker._margin_available)
                                    hist_dict["long_exposure"].append(
                                        broker._long_exposure
                                    )
                                    hist_dict["short_exposure"].append(
                                        broker._short_exposure
                                    )
                                    hist_dict["long_unrealised_pnl"].append(
                                        broker._long_unrealised_pnl
                                    )
                                    hist_dict["short_unrealised_pnl"].append(
                                        broker._short_unrealised_pnl
                                    )
                                    hist_dict["long_pnl"].append(
                                        broker._long_realised_pnl
                                    )
                                    hist_dict["short_pnl"].append(
                                        broker._short_realised_pnl
                                    )
                                    hist_dict["open_interest"].append(
                                        broker._open_interest
                                    )
                                    # TODO - check timezone below
                                    hist_dict["time"].append(datetime.now(timezone.utc))

                                # Dump history file to pickle
                                with open(f".paper_broker_hist", "wb") as file:
                                    pickle.dump(self._broker_histories, file)

                            # Go to sleep until next update
                            sleep_time = self._timestep.total_seconds() - (
                                (time.time() - deploy_time)
                                % self._timestep.total_seconds()
                            )
                            if int(self._verbosity) > 0:
                                print(
                                    f"AutoTrader sleeping until next update at {datetime.now()+timedelta(seconds=sleep_time)}."
                                )
                            time.sleep(sleep_time)

                            # Check if instance file still exists
                            instance_file_exists = self._check_instance_file(
                                instance_str
                            )

                        except KeyboardInterrupt:
                            print("\nKilling bot(s).")
                            instance_filepath = os.path.join(
                                self._home_dir, "active_bots", instance_str
                            )
                            try:
                                os.remove(instance_filepath)
                            except FileNotFoundError:
                                print(f"Intance file '{instance_str}' already deleted.")
                            break

            elif self._run_mode.lower() == "periodic":
                # Trading in periodic update mode
                if self._backtest_mode:
                    # Backtesting
                    self._check_bot_data()
                    start_range, end_range = self._bots_deployed[
                        0
                    ]._get_iteration_range()
                    for i in range(start_range, end_range):
                        # Update each bot with latest data to generate signal
                        for bot in self._bots_deployed:
                            bot._update(i=i)

                        # Update histories
                        for name, broker in self._brokers_dict.items():
                            hist_dict = self._broker_histories[name]
                            hist_dict["NAV"].append(broker._NAV)
                            hist_dict["equity"].append(broker._equity)
                            hist_dict["margin"].append(broker._margin_available)
                            hist_dict["open_interest"].append(broker._open_interest)
                            hist_dict["long_exposure"].append(broker._long_exposure)
                            hist_dict["short_exposure"].append(broker._short_exposure)
                            hist_dict["long_unrealised_pnl"].append(
                                broker._long_unrealised_pnl
                            )
                            hist_dict["short_unrealised_pnl"].append(
                                broker._short_unrealised_pnl
                            )
                            hist_dict["long_pnl"].append(broker._long_realised_pnl)
                            hist_dict["short_pnl"].append(broker._short_realised_pnl)
                            hist_dict["time"].append(broker._latest_time)

                else:
                    # Live trading
                    for bot in self._bots_deployed:
                        bot._update(i=-1)  # Process most recent signal

            # Run shutdown routines
            self.shutdown()

    @staticmethod
    def papertrade_snapshot(
        broker_picklefile: str = ".virtual_broker",
        history_picklefile: str = ".paper_broker_hist",
    ):
        """Prints a snapshot of the virtual broker from a single pickle. and
        returns the TradeAnalysis object."""
        broker = unpickle_broker(broker_picklefile)
        with open(history_picklefile, "rb") as file:
            broker_hist = pickle.load(file)

        # Extract relevant broker history

        # TODO - review functionality for multiple brokers: will need to pass
        # in as dict. Consider pickling self._brokers_dict

        results = TradeAnalysis(broker, broker_hist)
        at = AutoTrader()
        at.print_trade_results(results)
        return results

    def save_state(self):
        """Dumps the current AutoTrader instance to a pickle."""
        instance_id = self._get_instance_id(dir_name="pickled_instances")

        instance_name = (
            f"autotrader_instance_{instance_id}"
            if self._instance_str is None
            else self._instance_str
        )
        instance_file_exists = self._check_instance_file(
            instance_str=instance_name, dir_name="pickled_instances", live_check=False
        )

        write = "y"
        if instance_file_exists:
            # The file already exists, check to overwrite
            write = input(
                f"The instance file '{instance_name}' already "
                + "exists. Would you like to overwrite it? ([y]/n)  "
            )

        if "y" in write.lower():
            # Write to file
            try:
                filepath = f"pickled_instances/{instance_name}"
                with open(filepath, "wb") as file:
                    pickle.dump(self, file)
            except pickle.PicklingError:
                print("Error - cannot pickle this AutoTrader instance.")

    @staticmethod
    def load_state(instance_name, verbosity: int = 0):
        """Loads a pickled AutoTrader instance from file."""
        try:
            filepath = f"pickled_instances/{instance_name}"
            with open(filepath, "rb") as file:
                at = pickle.load(file)
            return at
        except Exception as e:
            print(f"Something went wrong while tring to load '{instance_name}'.")

            if verbosity > 0:
                print("Exception:", e)



import random
import ib_insync
import numpy as np
from autotrader.brokers.trading import Order, IsolatedPosition, Position
from autotrader.brokers.ib.utils import Utils


class Broker:
    """AutoTrader-InteractiveBrokers API interface.

    Attributes
    ----------
    utils : Utils
        The broker utilities.
    ib : ib_insync connection
        Used to query IB.
    account : str
        The active IB account.

    Notes
    -----
        - when closing a position using close_position(), if there are attached SL
          and/or TP orders, they must be closed manually using cancel_pending_order().
          Usually only one of the pair needs to be cancelled, and the other will too.
        - required signal_dict keys for different security types (eg. futures
          require symbol, exchange and contract_month)
        - The products localSymbol will always take precedence over the symbol.
          As such, it should be used as much as possible to avoid accidental
          actions.
    """

    def __init__(self, config: dict, utils: Utils = None) -> None:
        """Initialise AutoTrader-Interactive Brokers API interface.

        Parameters
        ----------
        config : dict
            The IB configuration dictionary. This can contain the host, port,
            clientID and read_only boolean flag.
        utils : Utils, optional
            Broker utilities class instance. The default is None.
        """
        self.utils = utils if utils is not None else Utils()

        self.host = config["host"] if "host" in config else "127.0.0.1"
        self.port = config["port"] if "port" in config else 7497
        self.client_id = (
            config["clientID"] if "clientID" in config else random.randint(0, 9999)
        )
        self.read_only = config["read_only"] if "read_only" in config else False
        self.account = config["account"] if "account" in config else ""

        self.ib = ib_insync.IB()
        self._check_connection()
        self.account = self._get_account() if self.account == "" else self.account

    def __repr__(self):
        return "AutoTrader-InteractiveBrokers interface"

    def __str__(self):
        return "AutoTrader-InteractiveBrokers interface"

    def get_NAV(self) -> float:
        """Returns the net asset/liquidation value of the account."""
        self._check_connection()
        summary = self.get_summary()
        return float(summary["NetLiquidation"]["value"])

    def get_balance(self) -> float:
        """Returns account balance."""
        self._check_connection()
        summary = self.get_summary()
        return float(summary["TotalCashValue"]["value"])

    def place_order(self, order: Order, **kwargs) -> None:
        """Disassembles order_details dictionary to place order.

        Parameters
        ----------
        order: Order
            The AutoTrader Order.

        Returns
        -------
        None
            Orders will be submitted to IB.
        """
        self._check_connection()

        # Call order to set order time
        order()

        if order.order_type == "market":
            self._place_market_order(order)
        elif order.order_type == "stop-limit":
            self._place_stop_limit_order(order)
        elif order.order_type == "limit":
            self._place_limit_order(order)
        elif order.order_type == "close":
            self._close_position(order)
        else:
            print("Order type not recognised.")

        self._refresh()

    def get_orders(self, instrument: str = None, **kwargs) -> dict:
        """Returns all pending orders (have not been filled) in the account.

        Parameters
        ----------
        instrument : str, optional
            The trading instrument's symbol. The default is None.

        Returns
        -------
        dict
            Pending orders for the requested instrument. If no instrument is provided,
            all pending orders will be returned.
        """
        self._check_connection()

        # Get all open trades
        open_trades = self.ib.openTrades()

        pending_orders = {}
        for trade in open_trades:
            trade_dict = trade.dict()
            contract = trade_dict["contract"]
            order_dict = trade_dict["order"].dict()
            order_status_dict = trade_dict["orderStatus"].dict()
            order_status = order_status_dict["status"]

            if order_status in ib_insync.OrderStatus.ActiveStates:
                # Order is still active (not yet filled)
                new_order = {}
                new_order["order_ID"] = order_dict["orderId"]
                new_order["order_type"] = order_dict["orderType"]
                new_order["order_stop_price"] = order_dict["auxPrice"]
                new_order["order_limit_price"] = order_dict["lmtPrice"]
                new_order["direction"] = 1 if order_dict["action"] == "BUY" else -1
                new_order["order_time"] = None
                new_order["instrument"] = contract.symbol
                new_order["size"] = order_dict["totalQuantity"]
                new_order["order_price"] = None
                new_order["take_profit"] = None
                new_order["take_distance"] = None
                new_order["stop_type"] = None
                new_order["stop_distance"] = None
                new_order["stop_loss"] = None
                new_order["related_orders"] = None
                new_order["granularity"] = None
                new_order["strategy"] = None

                if instrument is not None and contract.symbol == instrument:
                    pending_orders[new_order["order_ID"]] = Order._from_dict(new_order)
                elif instrument is None:
                    pending_orders[new_order["order_ID"]] = Order._from_dict(new_order)

        return pending_orders

    def cancel_order(self, order_id: int, **kwargs) -> list:
        """Cancels pending order by order ID.

        Parameters
        ----------
        order_id : int
            The ID of the order to be concelled.

        Returns
        -------
        list
            A list of the cancelled trades.

        """
        self._check_connection()

        open_trades = self.ib.openTrades()
        cancelled_trades = []
        for trade in open_trades:
            order = trade.order
            if order.orderId == order_id:
                cancel_trade = self.ib.cancelOrder(order)
                cancelled_trades.append(cancel_trade)

        return cancelled_trades

    def get_trades(self, instrument: str = None, **kwargs) -> dict:
        """Returns the open trades held by the account.

        Parameters
        ----------
        instrument : str, optional
            The trading instrument's symbol. The default is None.

        Returns
        -------
        dict
            The open trades.
        """
        self._check_connection()

        # Get all open trades
        all_open_trades = self.ib.openTrades()

        open_trades = {}
        for trade in all_open_trades:
            trade_dict = trade.dict()
            contract = trade_dict["contract"]
            order_dict = trade_dict["order"].dict()
            order_status_dict = trade_dict["orderStatus"].dict()
            order_status = order_status_dict["status"]

            if order_status == "Filled":
                # Trade order has been filled
                new_trade = {}
                new_trade["order_ID"] = order_dict["orderId"]
                new_trade["order_stop_price"] = order_dict["auxPrice"]
                new_trade["order_limit_price"] = order_dict["lmtPrice"]
                new_trade["direction"] = 1 if order_dict["action"] == "BUY" else -1
                new_trade["order_time"] = None
                new_trade["instrument"] = contract.symbol
                new_trade["size"] = order_dict["totalQuantity"]
                new_trade["order_price"] = None
                new_trade["entry_price"] = order_status_dict["lastFillPrice"]
                new_trade["order_type"] = None
                new_trade["take_profit"] = None
                new_trade["take_distance"] = None
                new_trade["stop_type"] = None
                new_trade["stop_distance"] = None
                new_trade["stop_loss"] = None
                new_trade["related_orders"] = None
                new_trade["granularity"] = None
                new_trade["strategy"] = None

                if instrument is not None and contract.symbol == instrument:
                    open_trades[new_trade["order_ID"]] = IsolatedPosition(new_trade)
                elif instrument is None:
                    open_trades[new_trade["order_ID"]] = IsolatedPosition(new_trade)

        return open_trades

    def get_trade_details(self, trade_ID: str, **kwargs) -> dict:
        """Returns the details of the trade specified by trade_ID.

        Parameters
        ----------
        trade_ID : str
            The ID of the trade.

        Returns
        -------
        dict
            The details of the trade.
        """
        raise NotImplementedError(
            "This method is not available, and will "
            + "be deprecated with a future release. Please use the "
            + "get_trades method instead."
        )

    def get_positions(self, instrument: str = None, **kwargs) -> dict:
        """Gets the current positions open on the account.

        Parameters
        ----------
        instrument : str, optional
            The trading instrument's symbol. This can be either the naive
            symbol, or the localSymbol. The default is None.

        Returns
        -------
        open_positions : dict
            A dictionary containing details of the open positions.

        Notes
        -----
        This function returns the position in an underlying product. If
        there are multiple contracts on the underlying, they will be
        appended to the returned Position objects' "contracts" and
        "portfolio_items" attributes.
        """

        def adjust_position(existing_pos, portfolio_item, pos_dict):
            # Position item already exists, append portfolio item
            existing_pos.long_units += pos_dict["long_units"]
            existing_pos.long_PL += pos_dict["long_PL"]
            existing_pos.short_units += pos_dict["short_units"]
            existing_pos.short_PL += pos_dict["short_PL"]
            existing_pos.net_position += pos_dict["net_position"]
            existing_pos.PL += pos_dict["PL"]
            existing_pos.contracts.append(portfolio_item.contract)
            existing_pos.portfolio_items.append(portfolio_item)

        self._check_connection()

        all_portfolio_items = self.ib.portfolio()
        open_positions = {}
        for item in all_portfolio_items:
            units = item.position
            pnl = item.unrealizedPNL
            pos_symbol = item.contract.symbol
            pos_dict = {
                "long_units": units if np.sign(units) > 0 else 0,
                "long_PL": pnl if np.sign(units) > 0 else 0,
                "short_units": abs(units) if np.sign(units) < 0 else 0,
                "short_PL": pnl if np.sign(units) < 0 else 0,
                "net_position": units,
                "PL": pnl,
                "contracts": [item.contract],
                "portfolio_items": [item],
                "instrument": pos_symbol,
            }

            symbol_match = instrument == pos_symbol  # Product symbol match
            localSymbol_match = (
                instrument == item.contract.localSymbol
            )  # localSymbol match
            unique_match = symbol_match or localSymbol_match

            if instrument is not None and unique_match:
                # The current item matches the request, append it
                key_symbol = (
                    item.contract.localSymbol
                    if localSymbol_match
                    else item.contract.symbol
                )

                pos_dict["instrument"] = key_symbol

                if pos_symbol in open_positions:
                    # Portfolio item already exists, append portfolio item
                    existing_pos = open_positions[pos_symbol]
                    adjust_position(existing_pos, item, pos_dict)
                else:
                    # New position
                    open_positions[key_symbol] = Position(**pos_dict)

            elif instrument is None:
                # Append all positions
                if pos_symbol in open_positions:
                    # Position item already exists, append portfolio item
                    existing_pos = open_positions[pos_symbol]
                    adjust_position(existing_pos, item, pos_dict)
                else:
                    # New position
                    open_positions[pos_symbol] = Position(**pos_dict)

        return open_positions

    def get_summary(self) -> dict:
        """Returns account summary."""
        self._check_connection()
        raw_summary = self.ib.accountSummary(self.account)
        summary = self.utils.accsum_to_dict(self.account, raw_summary)

        return summary

    def _get_historical_data(
        self, instrument: str, interval: str, from_time: str, to_time: str
    ):
        """Returns historical price data."""
        # self._check_connection()
        # self.ib.reqHistoricalData()
        raise NotImplementedError("This method is not available.")

    def _connect(self):
        """Connects from IB application."""
        self.ib.connect(
            host=self.host,
            port=self.port,
            clientId=self.client_id,
            readonly=self.read_only,
            account=self.account,
        )

    def _disconnect(self):
        """Disconnects from IB application."""
        self.ib.disconnect()

    def _check_connection(self):
        """Checks if there is an active connection to IB. If not, will
        attempt to reconnect.
        """
        self._refresh()
        connected = self.ib.isConnected()

        while not connected:
            try:
                # Try to connect
                self.ib = ib_insync.IB()
                self._connect()
            except:
                print("Connection to IB failed... trying to reconnect.")
                # Connection failed, increment client ID
                self.client_id = random.randint(0, 9999)

                # Sleep for a little while
                self.ib.sleep(10)

            # Update connection status
            connected = self.ib.isConnected()

    def _refresh(self):
        """Refreshes IB session events."""
        self.ib.sleep(0)

    def _get_account(
        self,
    ):
        """Returns the first managed account."""
        self._check_connection()
        accounts = self.ib.managedAccounts()
        return accounts[0]

    def _close_position(self, order: Order, **kwargs):
        """Closes open position of symbol by placing opposing market order.

        Warning
        -------
        If the order instrument is for an underlying product, all contracts
        held attributed to the underlying will be closed.
        """
        self._check_connection()

        symbol = (
            order.localSymbol if order.localSymbol is not None else order.instrument
        )
        positions = self.get_positions(instrument=symbol)
        position = positions[symbol]

        for item in position.portfolio_items:
            # Place opposing market order
            item_units = item.position
            action = "BUY" if item_units < 0 else "SELL"
            units = abs(item_units)
            IB_order = ib_insync.MarketOrder(action, units)
            contract = item.contract
            self.ib.qualifyContracts(contract)
            self.ib.placeOrder(contract, IB_order)

    def _place_market_order(self, order: Order):
        """Places a market order."""
        self._check_connection()

        # Build contract
        contract = self.utils.build_contract(order)

        # Create market order
        action = "BUY" if order.direction > 0 else "SELL"
        units = abs(order.size)
        market_order = ib_insync.MarketOrder(
            action, units, orderId=self.ib.client.getReqId(), transmit=False
        )

        # Attach SL and TP orders
        orders = self._attach_auxiliary_orders(order, market_order)

        # Submit orders
        self._process_orders(contract, orders)

    def _place_stop_limit_order(self, order):
        """Places stop-limit order."""
        self._check_connection()

        # Build contract
        contract = self.utils.build_contract(order)

        # Create stop limit order
        action = "BUY" if order.direction > 0 else "SELL"
        units = abs(order.size)
        lmtPrice = order.order_limit_price
        stopPrice = order.order_stop_price
        IBorder = ib_insync.StopLimitOrder(
            action,
            units,
            lmtPrice,
            stopPrice,
            orderId=self.ib.client.getReqId(),
            transmit=False,
        )

        # Attach SL and TP orders
        orders = self._attach_auxiliary_orders(order, IBorder)

        # Submit orders
        self._process_orders(contract, orders)

    def _place_limit_order(self, order):
        """Places limit order."""
        self._check_connection()

        # Build contract
        contract = self.utils.build_contract(order)

        action = "BUY" if order.direction > 0 else "SELL"
        units = abs(order.size)
        lmtPrice = order.order_limit_price
        IBorder = ib_insync.LimitOrder(
            action, units, lmtPrice, orderId=self.ib.client.getReqId(), transmit=False
        )

        # Attach SL and TP orders
        orders = self._attach_auxiliary_orders(order, IBorder)

        # Submit orders
        self._process_orders(contract, orders)

    def _attach_auxiliary_orders(
        self, order: Order, parent_order: ib_insync.order
    ) -> list:
        orders = [parent_order]

        # TP order
        if order.take_profit is not None:
            takeProfit_order = self._create_take_profit_order(
                order, parent_order.orderId
            )
            orders.append(takeProfit_order)

        # SL order
        if order.stop_loss is not None:
            stopLoss_order = self._create_stop_loss_order(order, parent_order.orderId)
            orders.append(stopLoss_order)

        return orders

    def _process_orders(self, contract: ib_insync.Contract, orders: list) -> None:
        """Processes a list of orders for a given contract."""
        self._check_connection()

        # Submit orders
        for i, order in enumerate(orders):
            if i == len(orders) - 1:
                # Final order; set transmit to True
                order.transmit = True
            else:
                order.transmit = False
            self.ib.placeOrder(contract, order)

    def _convert_to_oca(
        self, orders: list, oca_group: str = None, oca_type: int = 1
    ) -> list:
        """Converts a list of Orders to One Cancels All group of orders.

        Parameters
        ----------
        orders : list
            A list of orders.

        Returns
        -------
        oca_orders : list
            The orders modified to be in a OCA group.
        """
        self._check_connection()

        if oca_group is None:
            oca_group = f"OCA_{self.ib.client.getReqId()}"

        oca_orders = self.ib.oneCancelsAll(orders, oca_group, oca_type)
        return oca_orders

    def _create_take_profit_order(self, order: Order, parentId: int):
        """Constructs a take profit order."""
        quantity = order.size
        takeProfitPrice = order.take_profit
        action = "BUY" if order.direction < 0 else "SELL"
        takeProfit_order = ib_insync.LimitOrder(
            action,
            quantity,
            takeProfitPrice,
            orderId=self.ib.client.getReqId(),
            transmit=False,
            parentId=parentId,
        )
        return takeProfit_order

    def _create_stop_loss_order(self, order: Order, parentId: int):
        """Constructs a stop loss order."""
        # TODO - add support for trailing SL
        quantity = order.size
        stopLossPrice = order.stop_loss
        action = "BUY" if order.direction < 0 else "SELL"
        stopLoss_order = ib_insync.StopOrder(
            action,
            quantity,
            stopLossPrice,
            orderId=self.ib.client.getReqId(),
            transmit=True,
            parentId=parentId,
        )
        return stopLoss_order



import os
import numpy as np
import pandas as pd
from decimal import Decimal
from datetime import datetime


class BrokerUtils:
    def __init__(self):
        pass

    def __repr__(self):
        return "AutoTrader Broker Utilities"

    def __str__(self):
        return "AutoTrader Broker Utilities"

    def response_to_df(self, response: pd.DataFrame):
        """Function to convert api response into a pandas dataframe."""
        candles = response.body["candles"]
        times = []
        close_price, high_price, low_price, open_price = [], [], [], []

        for candle in candles:
            times.append(candle.time)
            close_price.append(float(candle.mid.c))
            high_price.append(float(candle.mid.h))
            low_price.append(float(candle.mid.l))
            open_price.append(float(candle.mid.o))

        dataframe = pd.DataFrame(
            {
                "Open": open_price,
                "High": high_price,
                "Low": low_price,
                "Close": close_price,
            }
        )
        dataframe.index = pd.to_datetime(times)

        return dataframe

    def truncate(self, f: float, n: int):
        """Truncates a float f to n decimal places without rounding."""
        s = "{}".format(f)

        if "e" in s or "E" in s:
            return "{0:.{1}f}".format(f, n)
        i, p, d = s.partition(".")

        return ".".join([i, (d + "0" * n)[:n]])

    def get_pip_ratio(self, pair):
        """Function to return pip value ($/pip) of a given forex pair.
        If you are not trading FX, stop losses should only be provided
        by the stop loss price (rather than a distance) to avoid
        unexpected results.
        """
        if "JPY" in pair:
            pip_value = 1e-2
        else:
            pip_value = 1e-4

        return pip_value

    def get_size(
        self,
        instrument: str,
        amount_risked: float,
        price: float,
        HCF: float,
        stop_price: float = None,
        stop_distance: float = None,
    ) -> float:
        """Calculate position size based on account balance and risk profile."""
        if stop_price is None and stop_distance is None:
            # No stop loss being used, instead risk portion of account
            units = amount_risked / (HCF * price)

        else:
            # SL provided
            if stop_price is None:
                # Stop distance provided (assume FX)
                pip_value = self.get_pip_ratio(instrument)
                price_distance = stop_distance * pip_value
            else:
                price_distance = abs(price - stop_price)

            # Calculate units
            if price_distance == 0:
                units = 0
            else:
                quote_risk = amount_risked / HCF
                units = quote_risk / price_distance

        return units

    def check_precision(self, pair, original_stop, original_take):
        """Modify stop/take based on pair for required ordering precision."""
        if pair[-3:] == "JPY":
            N = 3
        else:
            N = 5

        take_price = float(self.truncate(original_take, N))
        stop_price = float(self.truncate(original_stop, N))

        return stop_price, take_price

    def interval_to_seconds(self, interval):
        """Converts the interval to time in seconds."""
        letter = interval[0]

        if len(interval) > 1:
            number = float(interval[1:])
        else:
            number = 1

        conversions = {"S": 1, "M": 60, "H": 60 * 60, "D": 60 * 60 * 24}

        my_int = conversions[letter] * number

        return my_int

    def write_to_order_summary(self, order, filepath: str):
        """Writes order details to summary file."""
        # Check if file exists already, if not, create
        if not os.path.exists(filepath):
            f = open(filepath, "w")
            f.write(
                "order time, strategy, granularity, order_type, instrument, order_size, "
            )
            f.write("trigger_price, stop_loss, take_profit\n")
            f.close()

        order_time = order.order_time
        strategy = order.strategy
        order_type = order.order_type
        instrument = order.instrument
        size = order.size
        trigger_price = order.order_price
        stop_loss = order.stop_loss
        take_profit = order.take_profit
        granularity = order.granularity

        f = open(filepath, "a")
        f.write(
            "{}, {}, {}, {}, {}, {}, {}, {}, {}\n".format(
                order_time,
                strategy,
                granularity,
                order_type,
                instrument,
                size,
                trigger_price,
                stop_loss,
                take_profit,
            )
        )
        f.close()

    def check_dataframes(self, df_1: pd.DataFrame, df_2: pd.DataFrame):
        """Checks dataframe lengths and corrects if necessary."""
        if len(df_1) < len(df_2):
            new_df_1 = self.fix_dataframe(df_2, df_1)
            new_df_2 = df_2
        elif len(df_1) > len(df_2):
            new_df_2 = self.fix_dataframe(df_1, df_2)
            new_df_1 = df_1

        else:
            new_df_1 = df_1
            new_df_2 = df_2

        return new_df_1, new_df_2

    def fix_dataframe(self, df1: pd.DataFrame, df2: pd.DataFrame) -> pd.DataFrame:
        """Ensures that the quote data and data dataframes are the same
        lenght.
        """
        # Would be good to check which one is shorter, which is longer, then
        # return both with corrections

        i1 = list(df1.index)
        i2 = list(df2.index)
        new_indices = list(set(i1) - set(i2))

        new_df = df2

        for index in new_indices:

            df_row = df1.copy()[df1.index == index]

            df_row.Open = None
            df_row.High = None
            df_row.Low = None
            df_row.Close = None

            new_df = new_df.append(df_row)

        new_df = new_df.sort_index()
        new_df = new_df.interpolate()

        return new_df


class OrderBook:
    def __init__(self, instrument, initial_state):
        self.instrument = instrument
        self.bids = None
        self.asks = None
        self._midprice = None
        self._spread = None

        # Initialise from initial state
        self.bids = pd.DataFrame(initial_state["bids"]).astype(float)
        self.asks = pd.DataFrame(initial_state["asks"]).astype(float)

        # Sort quotes
        self.bids.sort_values(by="price", ascending=False, inplace=True)
        self.asks.sort_values(by="price", ascending=True, inplace=True)

        # Calculate spread and midprice
        spread = float(self.asks.price.min()) - float(self.bids.price.max())
        midprice = (float(self.asks.price.min()) + float(self.bids.price.max())) / 2

        # Quantize
        # TODO - use ticksize and step size to quantize
        ref = Decimal(str(self.bids["price"][0]))
        self.spread = float(Decimal(spread).quantize(ref))
        self.midprice = float(Decimal(midprice).quantize(ref))

    def __repr__(self):
        return f"{self.instrument} Order Book snapshot"

    @property
    def midprice(self):
        return self._midprice

    @midprice.setter
    def midprice(self, value):
        self._midprice = value

    @property
    def spread(self):
        return self._spread

    @spread.setter
    def spread(self, value):
        self._spread = value



import os
from macd_strategy import SimpleMACD
from limit_strategy import LimitStrategy
from autotrader.autotrader import AutoTrader


def test_macd_backtest():
    config = {
        "NAME": "MACD Strategy",
        "MODULE": "macd_strategy",
        "CLASS": "SimpleMACD",
        "INTERVAL": "4h",
        "PERIOD": 300,
        "RISK_PC": 1.5,
        "SIZING": "risk",
        "PARAMETERS": {
            "ema_period": 200,
            "MACD_fast": 5,
            "MACD_slow": 19,
            "MACD_smoothing": 9,
            "RR": 1.5,
        },
        "WATCHLIST": ["EUR_USD"],
    }
    home_dir = os.path.abspath(os.path.dirname(__file__))

    at = AutoTrader()
    at.configure(verbosity=1, show_plot=False, mode="periodic")
    at.add_strategy(config_dict=config, strategy=SimpleMACD)
    at.plot_settings(show_cancelled=True)
    at.add_data(
        {"EUR_USD": "EUR_USD_H4.csv"}, data_directory=os.path.join(home_dir, "data")
    )
    at.backtest(start="1/1/2021", end="1/1/2022")
    at.virtual_account_config(
        initial_balance=1000,
        leverage=30,
        spread=0.5 * 1e-4,
        commission=0.005,
        hedging=True,
    )
    at.run()
    bot = at.get_bots_deployed()
    bt_results = at.trade_results.summary()

    # Test backtest results
    assert bt_results["no_trades"] == 70, (
        "Incorrect number of trades " + "(single instrument backtest)"
    )
    assert round(bt_results["ending_balance"], 3) == 903.857, (
        "Incorrect " + "ending balance (single instrument backtest)"
    )


def test_multibot_macd_backtest():
    config = {
        "NAME": "MACD Strategy",
        "MODULE": "macd_strategy",
        "CLASS": "SimpleMACD",
        "INTERVAL": "4h",
        "PERIOD": 300,
        "RISK_PC": 1.5,
        "SIZING": "risk",
        "PARAMETERS": {
            "ema_period": 200,
            "MACD_fast": 5,
            "MACD_slow": 19,
            "MACD_smoothing": 9,
            "RR": 1.5,
        },
        "WATCHLIST": ["EUR_USD", "EUR_USD2"],
    }
    home_dir = os.path.abspath(os.path.dirname(__file__))

    at = AutoTrader()
    at.configure(verbosity=0, show_plot=False, mode="periodic")
    at.add_strategy(config_dict=config, strategy=SimpleMACD)
    at.plot_settings(show_cancelled=False)
    at.add_data(
        {"EUR_USD": "EUR_USD_H4.csv", "EUR_USD2": "EUR_USD_H4.csv"},
        data_directory=os.path.join(home_dir, "data"),
    )
    at.backtest(start="1/1/2021", end="1/1/2022")
    at.virtual_account_config(
        initial_balance=1000,
        leverage=30,
        spread=0.5 * 1e-4,
        commission=0.005,
        hedging=True,
    )
    at.run()
    bt_results = at.trade_results.summary()

    assert bt_results["no_trades"] == 134, (
        "Incorrect number of trades" + " (multi-instrument backtest)"
    )
    assert round(bt_results["ending_balance"], 3) == 838.037, (
        "Incorrect " + "ending balance (multi-instrument backtest)"
    )


def test_limit_backtest():
    config = {
        "NAME": "Limit Order Strategy",
        "CLASS": "LimitStrategy",
        "INTERVAL": "4h",
        "PERIOD": 50,
        "PARAMETERS": {},
        "SIZING": 100,
        "WATCHLIST": ["EUR_USD"],
    }
    home_dir = os.path.abspath(os.path.dirname(__file__))

    at = AutoTrader()
    at.configure(verbosity=1, show_plot=False, mode="continuous", update_interval="4h")
    at.add_strategy(config_dict=config, strategy=LimitStrategy)
    at.plot_settings(show_cancelled=True)
    at.add_data(
        {"EUR_USD": "EUR_USD_H4.csv"}, data_directory=os.path.join(home_dir, "data")
    )
    at.backtest(start="1/1/2021", end="1/3/2021")
    at.virtual_account_config(
        initial_balance=1000, leverage=30, spread=0.5 * 1e-4, commission=0.005
    )
    at.run()

    bt_results = at.trade_results.summary()

    # Test backtest results
    assert bt_results["no_trades"] == 2, (
        "Incorrect number of trades " + "(limit order backtest)"
    )
    assert round(bt_results["ending_balance"], 3) == 1000.988, (
        "Incorrect " + "ending balance (limit order backtest)"
    )


def test_margin_call_backtest():
    config = {
        "NAME": "Limit Order Strategy",
        "CLASS": "LimitStrategy",
        "INTERVAL": "4h",
        "PERIOD": 50,
        "PARAMETERS": {},
        "SIZING": 10000,
        "WATCHLIST": ["EUR_USD"],
    }
    home_dir = os.path.abspath(os.path.dirname(__file__))

    at = AutoTrader()
    at.configure(verbosity=1, show_plot=False, mode="continuous", update_interval="4h")
    at.add_strategy(config_dict=config, strategy=LimitStrategy)
    at.plot_settings(show_cancelled=True)
    at.add_data(
        {"EUR_USD": "EUR_USD_H4.csv"}, data_directory=os.path.join(home_dir, "data")
    )
    at.backtest(start="1/1/2021", end="1/3/2021")
    at.virtual_account_config(
        initial_balance=1000,
        leverage=30,
        margin_call_fraction=0.6,
        spread=0.5 * 1e-4,
        commission=0.005,
    )
    at.run()

    bt_results = at.trade_results.summary()

    # Test backtest results
    assert bt_results["no_trades"] == 4, (
        "Incorrect number of trades " + "(margin call backtest)"
    )
    assert round(bt_results["ending_balance"], 3) == 970.769, (
        "Incorrect " + "ending balance (margin call backtest)"
    )



import os
import time
import click
import shutil
import requests
import autotrader
from art import tprint


def print_banner():
    tprint("AutoTrader", font="tarty1")


def download_file(url):
    local_filename = url.split("/")[-1]
    with requests.get(url, stream=True) as r:
        r.raise_for_status()
        with open(local_filename, "wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)
    return local_filename


@click.group()
def cli():
    """AutoTrader command line interface."""
    pass


@click.command()
def version():
    """Shows the installed version number of AutoTrader."""
    print(autotrader.__version__)


@click.command()
@click.option(
    "-m",
    "--minimal",
    is_flag=True,
    default=False,
    help="Minimal directory initialisation.",
)
@click.option(
    "-s",
    "--strategies",
    help="The name of strategies to include in the initialised directory.",
)
@click.argument("name", default=".")
def init(minimal, strategies, name):
    """Initialises the directory NAME for trading with AutoTrader. If no
    directory NAME is provided, the current directory will be initialised.
    """
    print_banner()

    # Construct filepaths
    file_dir = os.path.dirname(os.path.abspath(__file__))
    data_dir = os.path.join(file_dir, "..", "data")
    if name != ".":
        # Initialise directory specified
        dir_name = name
        if not os.path.isdir(dir_name):
            # Directory doesn't exist yet - create it
            os.mkdir(dir_name)
    else:
        # Initialise current directory
        dir_name = os.path.abspath(os.getcwd())

    if minimal:
        # Run minimial directory initialisation
        pass

    else:
        # Run full directory initialisation
        # Check if config directory exists
        config_dir = os.path.join(dir_name, "config")
        if not os.path.isdir(config_dir):
            os.mkdir(config_dir)

        # Copy keys config
        keys_config_fp = os.path.join(data_dir, "keys.yaml")
        shutil.copyfile(keys_config_fp, os.path.join(config_dir, "keys.yaml"))

        # Check if strategy directory exists
        strategy_dir = os.path.join(dir_name, "strategies")
        if not os.path.isdir(strategy_dir):
            # Strategy directory doesn't exist - create it
            os.mkdir(strategy_dir)

    # Add strategies
    if strategies is not None:
        for strategy in strategies.split(","):
            # Construct urls
            if strategy == "template":
                # Get strategy template from main repo
                urls = {
                    "strategies": "https://raw.githubusercontent.com/"
                    + "kieran-mackle/AutoTrader/main/templates/strategy.py",
                    "config": None,
                }

            elif strategy == "config":
                # Get strategy config file
                urls = {
                    "strategies": None,
                    "config": "https://raw.githubusercontent.com/"
                    + "kieran-mackle/AutoTrader/main/templates/strategy_config.yaml",
                }

            else:
                # Get from demo repo
                urls = {
                    "strategies": "https://raw.githubusercontent.com/kieran-mackle/"
                    + f"autotrader-demo/main/strategies/{strategy}.py",
                    "config": "https://raw.githubusercontent.com/kieran-mackle/"
                    + f"autotrader-demo/main/config/{strategy}.yaml",
                }

            for dir in ["strategies", "config"]:
                try:
                    # Download
                    filename = download_file(urls[dir])

                    # Move to appropriate directory
                    if minimal:
                        move_to = os.path.join(dir_name, filename)
                    else:
                        move_to = os.path.join(dir_name, dir, filename)
                    os.rename(filename, move_to)

                except:
                    pass

    # Print completion message
    click.echo("AutoTrader initialisation complete.")


@click.command()
def demo():
    """Runs a demo backtest in AutoTrader."""
    # Download the strategy file and data
    print("Loading demo files...")
    strat_filename = download_file(
        "https://raw.githubusercontent.com/kieran-mackle/"
        + "AutoTrader/main/tests/macd_strategy.py"
    )
    data_filename = download_file(
        "https://raw.githubusercontent.com/kieran-mackle/"
        + "AutoTrader/main/tests/data/EUR_USD_H4.csv"
    )
    print("  Done.")

    # Run backtest


@click.command()
@click.option("-p", "--port", default=8009, help="The port to serve data to.")
@click.option(
    "-n", "--nav", default=None, help="The reference NAV to use for relative PnL."
)
@click.option(
    "-f", "--file", help="The pickle file containing a virtual broker instance."
)
@click.option("-b", "--broker", help="The name of the broker to connect to.")
@click.option("-e", "--environment", default="paper", help="The trading environment.")
def monitor(port, nav, file, broker, environment):
    """Monitors a broker and serves the information
    to a prometheus database."""
    broker_name = broker

    def start_server(port):
        """Starts the http server for Prometheus."""
        start_http_server(port)
        print(f"Server started on port {port}.")

    def get_broker(broker):
        """Returns the broker object."""
        if file is not None:
            # Unpickle latest broker instance
            broker = unpickle_broker(picklefile=picklepath)
        elif broker is not None:
            # Use existing broker instance
            pass
        elif broker_name is not None:
            # Create broker instance
            print(f"Connecting to {broker_name}...")
            at = autotrader.AutoTrader()
            at.configure(broker=broker_name, environment=environment, verbosity=0)
            broker = at.run()
            print("  Done.")
        return broker

    # Import packages
    from autotrader.utilities import unpickle_broker
    from prometheus_client import start_http_server, Gauge

    print_banner()

    # Unpack inputs
    ref_nav = nav
    picklepath = file

    if picklepath is not None:
        # Check picklefile exists
        if not os.path.exists(picklepath):
            raise Exception(f"\nPicklefile '{file}' does not exist!")
        else:
            print(f"Monitoring {file}.")

    # Set up instrumentation
    nav_gauge = Gauge("nav_gauge", "Net Asset Value gauge.")
    abs_PnL_gauge = Gauge("abs_pnl_gauge", "Absolute ($) PnL gauge.")
    rel_PnL_gauge = Gauge("rel_pnl_gauge", "Relative (%) PnL gauge.")
    pos_gauge = Gauge("pos_gauge", "Number of open positions gauge.")
    total_exposure_gauge = Gauge("total_exposure_gauge", "Total exposure gauge.")
    net_exposure_gauge = Gauge("net_exposure_gauge", "Total exposure gauge.")
    leverage_gauge = Gauge("leverage_gauge", "Total leverage gauge.")

    # Start up the server to expose the metrics
    try:
        start_server(port)
    except OSError:
        # Kill existing server
        from psutil import process_iter
        from signal import SIGKILL  # or SIGTERM

        for proc in process_iter():
            for conns in proc.connections(kind="inet"):
                if conns.laddr.port == port:
                    proc.send_signal(SIGKILL)  # or SIGKILL

        # Start server
        start_server(port)

    # Begin loop
    broker = None
    while True:
        try:
            # Get broker object
            broker = get_broker(broker)

            # Query broker
            nav = broker.get_NAV()
            if ref_nav is None:
                ref_nav = nav
            positions = broker.get_positions()
            pnl = nav - ref_nav
            rel_pnl = pnl / ref_nav

            # Calculate total exposure
            total_exposure = 0
            net_exposure = 0
            for instrument, position in positions.items():
                total_exposure += abs(position.notional)
                net_exposure += position.direction * position.notional

            # Calculate leverage
            leverage = total_exposure / nav

            # Update Prometheus server
            nav_gauge.set(nav)
            abs_PnL_gauge.set(pnl)
            rel_PnL_gauge.set(rel_pnl)
            pos_gauge.set(len(positions))
            total_exposure_gauge.set(total_exposure)
            net_exposure_gauge.set(net_exposure)
            leverage_gauge.set(leverage)

            # Sleep
            time.sleep(5)

        except KeyboardInterrupt:
            print("\n\nStopping monitoring.")
            break

        except Exception as e:
            # Unexpected exception, sleep briefly
            print(e)
            time.sleep(3)


@click.command()
@click.argument("pickle")
def snapshot(pickle):
    """Prints a snapshot of the trading account of a broker PICKLE instance
    file."""
    print_banner()
    autotrader.AutoTrader.papertrade_snapshot(pickle)
    print("")


# Add commands to CLI
cli.add_command(version)
cli.add_command(init)
cli.add_command(monitor)
cli.add_command(snapshot)
# cli.add_command(demo)



import pandas as pd
from decimal import Decimal
from datetime import datetime
from autotrader import AutoData
from dydx3 import Client, constants
from autotrader.brokers.broker_utils import OrderBook
from autotrader.brokers.dydx.utils import Utils, BrokerUtils
from autotrader.brokers.trading import Order, Position, Trade


class Broker:
    def __init__(self, config: dict, utils: BrokerUtils = None) -> None:
        """AutoTrader Broker Class constructor."""
        self._utils = utils if utils is not None else Utils()

        # Unpack config to obtain STARK and API keys
        client = Client(
            host="http://localhost:8080", eth_private_key=config["ETH_PRIV_KEY"]
        )
        STARK_KEYS = client.onboarding.derive_stark_key(config["ETH_ADDRESS"])
        API_KEY = client.onboarding.recover_default_api_key_credentials(
            config["ETH_ADDRESS"]
        )

        # Connect to dYdX API
        self.api = Client(
            host="https://api.dydx.exchange",
            api_key_credentials=API_KEY,
            stark_private_key=STARK_KEYS["private_key"],
            stark_public_key=STARK_KEYS["public_key"],
            stark_public_key_y_coordinate=STARK_KEYS["public_key_y_coordinate"],
            eth_private_key=config["ETH_PRIV_KEY"],
            default_ethereum_address=config["ETH_ADDRESS"],
        )

        # Create AutoData instance
        self.autodata = AutoData(data_source="dydx")

    def __repr__(self):
        return "AutoTrader-dYdX interface"

    def __str__(self):
        return "AutoTrader-dYdX interface"

    def get_NAV(self) -> float:
        """Returns the net asset/liquidation value of the account."""
        account = self._get_account()
        return float(account["equity"])

    def get_balance(self) -> float:
        """Returns account balance."""
        return self.get_NAV()

    def place_order(self, order: Order, **kwargs) -> None:
        """Disassemble order_details dictionary to place order."""
        # TODO - build in checking of constants.__dict__ for instrument

        # Call order to set order time
        order()

        # Extract information for submission to dydx
        order = self._check_order_precision(order)
        side = "BUY" if order.direction > 0 else "SELL"
        order_type = order.order_type.upper()
        expiration = int((pd.Timedelta("30days") + datetime.now()).timestamp())
        order_price = (
            str(order.order_limit_price)
            if order.order_limit_price is not None
            else None
        )
        trigger_price = (
            str(order.order_stop_price) if order.order_stop_price is not None else None
        )
        position_id = self._get_account()["positionId"]
        limit_fee = order.limit_fee

        kwargs = {}
        if trigger_price is not None:
            kwargs["trigger_price"] = trigger_price
        if order_type == "MARKET":
            kwargs["time_in_force"] = "IOC"

        # TODO - allow more kwargs to be parsed from order

        if order_price is None:
            # Create order price
            midprice = self.autodata.L2(order.instrument).midprice
            multiple = 1.05 if side == "BUY" else 0.95
            order_price = (Decimal(multiple) * midprice).quantize(midprice)

        # Submit order to dydx
        order = self.api.private.create_order(
            position_id=position_id,
            market=order.instrument,
            side=side,
            order_type=order_type,
            post_only=order.post_only,
            size=str(order.size),
            price=str(order_price),
            limit_fee=limit_fee,
            expiration_epoch_seconds=expiration,
            **kwargs,
        )

        return self._native_order(order.data["order"])

    def get_orders(self, instrument: str = None, **kwargs) -> dict:
        """Returns all pending orders (have not been filled) in the account.

        kwargs can include status, side, type, limit, createdBeforeOrAt and returnLatestOrders. See
        https://docs.dydx.exchange/?python#get-orders for more details.
        """
        orders = self.api.private.get_orders(market=instrument, **kwargs)
        orders = self._conver_order_list(orders.data["orders"])
        return orders

    def cancel_order(self, order_id: int, **kwargs) -> None:
        """Cancels order by order ID."""
        cancelled_order = self.api.private.cancel_order(order_id)
        return self._native_order(cancelled_order.data["cancelOrder"])

    def cancel_all_orders(self, instrument: str = None, **kwargs):
        cancelled_orders = self.api.private.cancel_all_orders(market=instrument)
        return cancelled_orders

    def get_trades(self, instrument: str = None, **kwargs) -> dict:
        """Returns the open trades held by the account."""
        fills = self.api.private.get_fills(market=instrument)
        trades = self._convert_fills(fills.data["fills"])
        return trades

    def get_trade_details(self, trade_ID: str) -> dict:
        """Returns the details of the trade specified by trade_ID."""
        raise NotImplementedError(
            "This method is not available, and will "
            + "be deprecated with a future release. Please use the "
            + "get_trades method instead."
        )

    def get_positions(self, instrument: str = None, **kwargs) -> dict:
        """Gets the current positions open on the account.

        Parameters
        ----------
        instrument : str, optional
            The trading instrument name (symbol). The default is None.

        Returns
        -------
        open_positions : dict
            A dictionary containing details of the open positions.
        """
        status = kwargs["status"].upper() if "status" in kwargs else "OPEN"
        positions = self.api.private.get_positions(market=instrument, status=status)
        positions = self._convert_position_list(positions.data["positions"])
        return positions

    def get_markets(self):
        # Get Markets
        markets = self.api.public.get_markets()
        market_df = pd.DataFrame(markets.data["markets"])
        return market_df

    def get_orderbook(self, instrument: str) -> OrderBook:
        # Get Orderbook
        orderbook = self.autodata.L2(instrument=instrument)
        return orderbook

    def get_market_stats(self):
        # Get Market Statistics
        market_statistics = self.api.public.get_stats(
            market=constants.MARKET_ADA_USD,
            days=1,
        )
        return market_statistics

    def get_funding(self, dtime):
        # Funding Data
        historical_funding = self.api.public.get_historical_funding(
            market=constants.MARKET_BTC_USD, effective_before_or_at=dtime
        )

        funding_df = pd.DataFrame(historical_funding.data["historicalFunding"])
        funding_df["rate"] = pd.to_numeric(funding_df["rate"], errors="coerce")
        funding_df["price"] = pd.to_numeric(funding_df["price"], errors="coerce")
        funding_df["effectiveAt"] = pd.to_datetime(
            funding_df["effectiveAt"], format="%Y-%m-%dT%H:%M:%S.%f"
        )
        funding_df["rate"] = funding_df["rate"] * 100

        return funding_df

    def get_candles(
        self,
    ):
        # Candlestick Data
        candles = self.api.public.get_candles(
            market=constants.MARKET_BTC_USD,
            resolution="1MIN",
        )
        candles = pd.DataFrame(candles.data["candles"])
        candles.apply(pd.to_numeric, errors="ignore").info()
        return candles

    def _get_account(self, eth_address: str = None):
        if eth_address is None:
            eth_address = self.api.default_address
        account = self.api.private.get_account(eth_address)
        return account.data["account"]

    def _get_market(self, instrument: str):
        """Returns the dydx market constant from an instrument."""
        pass

    def _native_order(self, dydx_order):
        """Helper method to convert a dydx order into a native AutoTrader Order."""

        direction = 1 if dydx_order["side"] == "BUY" else -1
        order_limit_price = (
            float(dydx_order["price"]) if dydx_order["price"] is not None else None
        )
        order_stop_price = (
            float(dydx_order["triggerPrice"])
            if dydx_order["triggerPrice"] is not None
            else None
        )
        order = Order(
            instrument=dydx_order["market"],
            order_type=dydx_order["type"],
            status=dydx_order["status"].lower(),
            id=dydx_order["id"],
            direction=direction,
            size=float(dydx_order["size"]),
            order_limit_price=order_limit_price,
            order_stop_price=order_stop_price,
        )
        return order

    def _conver_order_list(self, order_list):
        orders = {}
        for order in order_list:
            native_order = self._native_order(order)
            orders[native_order.id] = native_order
        return orders

    def _native_position(self, dydx_position):
        """Converts a dydx position to a native AutoTrader Position."""
        if dydx_position["side"] == "SHORT":
            position_units = {
                "short_units": float(dydx_position["size"]),
                "short_PL": dydx_position["unrealizedPnl"],
            }
        else:
            position_units = {
                "long_units": float(dydx_position["size"]),
                "long_PL": dydx_position["unrealizedPnl"],
            }

        native_position = Position(
            instrument=dydx_position["market"],
            net_position=float(dydx_position["size"]),
            PL=dydx_position["unrealizedPnl"],
            entry_price=float(dydx_position["entryPrice"]),
            **position_units,
        )
        return native_position

    def _convert_position_list(self, dydx_position_list):
        positions = {}
        for position in dydx_position_list:
            positions[position["market"]] = self._native_position(position)
        return positions

    def _native_trade(self, dydx_fill: dict):
        """Converts a dydx fill to a native AutoTrader trade."""
        direction = 1 if dydx_fill["side"] == "BUY" else -1

        native_trade = Trade(
            instrument=dydx_fill["market"],
            order_price=None,
            order_time=None,
            order_type=dydx_fill["type"].lower(),
            size=float(dydx_fill["size"]),
            fill_time=dydx_fill["createdAt"],
            fill_price=float(dydx_fill["price"]),
            fill_direction=direction,
            fee=dydx_fill["fee"],
            id=dydx_fill["id"],
            order_id=dydx_fill["orderId"],
        )

        return native_trade

    def _convert_fills(self, dydx_fills: list):
        """Converts an array of fills to a Trades dictionary."""
        trades = {}
        for trade in dydx_fills:
            trades[trade["id"]] = self._native_trade(trade)
        return trades

    def get_instrument_details(self, instrument):
        """Returns details of the instrument provided."""
        markets = self.api.public.get_markets()
        try:
            details = markets.data["markets"][instrument]
        except KeyError:
            raise Exception(f"The requested instrument '{instrument}' is invalid.")
        return details

    def _check_order_precision(self, order: Order):
        """Enforces that an order has an allowable precision for price and size."""
        details = self.get_instrument_details(order.instrument)
        stepsize = float(details["stepSize"])
        ticksize = float(details["tickSize"])
        order.size = Decimal(str(order.size)) - Decimal(str(order.size)) % Decimal(
            str(stepsize)
        )

        if order.order_type == "limit":
            order.order_limit_price = Decimal(str(order.order_limit_price)).quantize(
                Decimal(str(ticksize))
            )

        elif order.order_type == "stop-limit":
            order.order_limit_price = Decimal(str(order.order_limit_price)).quantize(
                Decimal(str(ticksize))
            )
            order.order_stop_price = Decimal(str(order.order_stop_price)).quantize(
                Decimal(str(ticksize))
            )

        return order



import os
from finta import TA
import autotrader.indicators as indicators
from autotrader.brokers.trading import Order


class SimpleMACD:
    """Simple MACD Strategy

    Rules
    ------
    1. Trade in direction of trend, as per 200EMA.
    2. Entry signal on MACD cross below/above zero line.
    3. Set stop loss at recent price swing.
    4. Target 1.5 take profit.
    """

    def __init__(self, parameters, data, instrument):
        """Define all indicators used in the strategy."""

        self.name = "MACD Trend Strategy"
        self.params = parameters
        self.instrument = instrument

        # Initial feature calculation
        self.calculate_features(data)

        # Construct indicators dict for plotting
        self.indicators = {
            "MACD (12/26/9)": {
                "type": "MACD",
                "macd": self.MACD.MACD,
                "signal": self.MACD.SIGNAL,
            },
            "EMA (200)": {"type": "MA", "data": self.ema},
        }

    def calculate_features(self, data):

        # Save data for other functions
        self.data = data

        # 200EMA
        self.ema = TA.EMA(data, self.params["ema_period"])

        # MACD
        self.MACD = TA.MACD(
            data,
            self.params["MACD_fast"],
            self.params["MACD_slow"],
            self.params["MACD_smoothing"],
        )
        self.MACD_CO = indicators.crossover(self.MACD.MACD, self.MACD.SIGNAL)
        self.MACD_CO_vals = indicators.cross_values(
            self.MACD.MACD, self.MACD.SIGNAL, self.MACD_CO
        )

        # Price swings
        self.swings = indicators.find_swings(data)

        # TODO - append data to self.indicators as it is created

    def generate_signal(self, data):
        """Define strategy to determine entry signals."""
        # Feature calculation
        self.calculate_features(data)
        i = -1  # Now just use latest data
        if (
            self.data.Close.values[i] > self.ema[i]
            and self.MACD_CO[i] == 1
            and self.MACD_CO_vals[i] < 0
        ):
            exit_dict = self.generate_exit_levels(signal=1, i=i)
            new_order = Order(
                direction=1,
                stop_loss=exit_dict["stop_loss"],
                take_profit=exit_dict["take_profit"],
            )

        elif (
            self.data.Close.values[i] < self.ema[i]
            and self.MACD_CO[i] == -1
            and self.MACD_CO_vals[i] > 0
        ):
            exit_dict = self.generate_exit_levels(signal=-1, i=i)
            new_order = Order(
                direction=-1,
                stop_loss=exit_dict["stop_loss"],
                take_profit=exit_dict["take_profit"],
            )

        else:
            new_order = Order()

        return new_order

    def generate_exit_levels(self, signal, i):
        """Function to determine stop loss and take profit levels."""
        stop_type = "limit"
        RR = self.params["RR"]

        if signal == 0:
            stop = None
            take = None
        else:
            if signal == 1:
                stop = self.swings.Lows[i]
                take = self.data.Close[i] + RR * (self.data.Close[i] - stop)
            else:
                stop = self.swings.Highs[i]
                take = self.data.Close[i] - RR * (stop - self.data.Close[i])

        exit_dict = {"stop_loss": stop, "stop_type": stop_type, "take_profit": take}

        return exit_dict


# if __name__ == "__main__":
from autotrader.autotrader import AutoTrader

config = {
    "NAME": "MACD Strategy",
    "MODULE": "macd_strategy_continuous",
    "CLASS": "SimpleMACD",
    "INTERVAL": "4h",
    "PERIOD": 300,
    "RISK_PC": 1.5,
    "SIZING": "risk",
    "PARAMETERS": {
        "ema_period": 200,
        "MACD_fast": 5,
        "MACD_slow": 19,
        "MACD_smoothing": 9,
        "RR": 1.5,
    },
    "WATCHLIST": ["EUR_USD"],
}
home_dir = os.getcwd()
data_dir = "/home/kmackle/Documents/AutoTrader/tests/data"

at = AutoTrader()
at.configure(verbosity=1, show_plot=True)  # , broker='ccxt:binanceusdm, ccxt:bitmex')
at.add_strategy(config_dict=config, strategy=SimpleMACD)
at.plot_settings(show_cancelled=False)
at.add_data(
    {"EUR_USD": "EUR_USD_H4.csv"},
    # data_directory=os.path.join(home_dir, 'data'),
    data_directory=data_dir,
)
at.backtest(
    start="1/1/2021",
    end="1/6/2021",
)
at.virtual_account_config(
    # initial_balance=1000, leverage=30, spread=0.5 * 1e-4, commission=0.005
    initial_balance=1000,
    leverage=30,
    spread=0,
    commission=0.0,
    # exchange='ccxt:binanceusdm',
    # tradeable_instruments=["EUR_USD"],
)
at.virtual_account_config(
    # initial_balance=1000, leverage=30, spread=0.5 * 1e-4, commission=0.005
    initial_balance=1000,
    leverage=30,
    spread=0,
    commission=0.0,
    # exchange='ccxt:bitmex',
    # tradeable_instruments=["EUR_USD"],
)
at.run()

# bot = at.get_bots_deployed()



import os
import sys
from getopt import getopt
from autotrader.utilities import read_yaml
from autotrader.comms.emailing import send_order_summary


def main(uo_dict: dict) -> None:
    email_dir = os.path.dirname(os.path.abspath(__file__))
    file_dir = os.path.join(email_dir, "../logfiles")
    file_path = os.path.join(file_dir, uo_dict["filename"])

    global_config = read_yaml(email_dir + "/../config/GLOBAL.yaml")
    host_email = global_config["EMAILING"]["HOST_ACCOUNT"]
    mailing_list = global_config["EMAILING"]["MAILING_LIST"]

    send_order_summary(file_path, mailing_list, host_email)


short_options = "f:"
long_options = ["file="]


if __name__ == "__main__":
    options, r = getopt(sys.argv[1:], short_options, long_options)

    # Defaults
    filename = None

    for opt, arg in options:
        if opt in ("-f", "--file"):
            filename = arg

    uo_dict = {"filename": filename}

    main(uo_dict)



import ib_insync
from datetime import datetime
from autotrader.brokers.trading import Order
from dateutil.relativedelta import relativedelta
from autotrader.brokers.broker_utils import BrokerUtils


class Utils(BrokerUtils):
    def __init__(self, **kwargs):
        pass

    def __repr__(self):
        return "AutoTrader-InteractiveBrokers Utilities"

    def __str__(self):
        return "AutoTrader-InteractiveBrokers Utilities"

    def check_response(self, response):
        """Checks API response for errors.

        Parameters
        ----------
        response : TYPE
            DESCRIPTION.

        Returns
        -------
        output : TYPE
            DESCRIPTION.
        """

        if response.status != 201:
            message = response.body["errorMessage"]
        else:
            message = "Success."

        output = {"Status": response.status, "Message": message}

        return output

    def check_precision(self, pair, price):
        """Modify a price based on required ordering precision for pair."""
        N = self.get_precision(pair)
        corrected_price = round(price, N)

        return corrected_price

    def get_precision(self, pair):
        """Returns the allowable precision for a given pair"""

        response = self.api.account.instruments(
            accountID=self.ACCOUNT_ID, instruments=pair
        )

        precision = response.body["instruments"][0].displayPrecision

        return precision

    def check_trade_size(self, pair, units):
        """Checks the requested trade size against the minimum trade size
        allowed for the currency pair."""
        response = self.api.account.instruments(
            accountID=self.ACCOUNT_ID, instruments=pair
        )
        # minimum_units = response.body['instruments'][0].minimumTradeSize
        trade_unit_precision = response.body["instruments"][0].tradeUnitsPrecision

        return round(units, trade_unit_precision)

    @staticmethod
    def build_contract(order: Order) -> ib_insync.contract.Contract:
        """Builds IB contract from the order details."""
        instrument = order.instrument
        security_type = order.secType

        # Get contract object
        contract_object = getattr(ib_insync, security_type)

        if security_type == "Stock":
            # symbol='', exchange='', currency=''
            exchange = order.exchange if order.exchange else "SMART"
            currency = order.currency if order.currency else "USD"
            contract = contract_object(
                symbol=instrument, exchange=exchange, currency=currency
            )

        elif security_type == "Options":
            raise NotImplementedError(
                f"Contract building for {security_type.lower()} trading is not supported yet."
            )

        elif security_type == "Future":
            # Requires order_details{'instrument', 'exchange', 'contract_month'}
            exchange = order.exchange if order.exchange else "GLOBEX"
            currency = order.currency if order.currency else "USD"
            contract_month = order.contract_month
            local_symbol = order.localSymbol if order.localSymbol else ""
            contract = contract_object(
                symbol=instrument,
                exchange=exchange,
                currency=currency,
                lastTradeDateOrContractMonth=contract_month,
                localSymbol=local_symbol,
            )

        elif security_type == "ContFuture":
            raise NotImplementedError(
                f"Contract building for {security_type.lower()} trading is not supported yet."
            )

        elif security_type == "Forex":
            # pair='', exchange='IDEALPRO', symbol='', currency='', **kwargs)
            exchange = order.exchange if order.exchange else "IDEALPRO"
            contract = contract_object(pair=instrument, exchange=exchange)

        elif security_type == "Index":
            raise NotImplementedError(
                f"Contract building for {security_type.lower()} trading is not supported yet."
            )

        elif security_type == "CFD":
            # symbol='', exchange='', currency='',
            exchange = order.exchange if order.exchange else "SMART"
            currency = order.currency if order.currency else "USD"
            contract = contract_object(
                symbol=instrument, exchange=exchange, currency=currency
            )

        elif security_type == "Commodity":
            raise NotImplementedError(
                f"Contract building for {security_type.lower()} trading is not supported yet."
            )
        elif security_type == "Bond":
            raise NotImplementedError(
                f"Contract building for {security_type.lower()} trading is not supported yet."
            )
        elif security_type == "FuturesOption":
            raise NotImplementedError(
                f"Contract building for {security_type.lower()} trading is not supported yet."
            )
        elif security_type == "MutualFund":
            raise NotImplementedError(
                f"Contract building for {security_type.lower()} trading is not supported yet."
            )
        elif security_type == "Warrant":
            raise NotImplementedError(
                f"Contract building for {security_type.lower()} trading is not supported yet."
            )
        elif security_type == "Bag":
            raise NotImplementedError(
                f"Contract building for {security_type.lower()} trading is not supported yet."
            )
        elif security_type == "Crypto":
            raise NotImplementedError(
                f"Contract building for {security_type.lower()} trading is not supported yet."
            )

        return contract

    @staticmethod
    def accsum_to_dict(account: str = None, data: list = None) -> dict:
        """Returns account summary list as a dictionary.

        Parameters
        ----------
        account : str
            DESCRIPTION.
        data : list
            DESCRIPTION.

        Returns
        -------
        out
            DESCRIPTION.
        """

        if account is None:
            account = "All"

        out = {}
        for av in data:
            if av.account == account:
                out[av.tag] = {
                    "value": av.value,
                    "currency": av.currency,
                    "modelCode": av.modelCode,
                }

        return out

    @staticmethod
    def positionlist_to_dict(positions: list) -> dict:
        """Returns position list as a dictionary.

        Parameters
        ----------
        positions : list
            DESCRIPTION.

        Returns
        -------
        dict
            DESCRIPTION.

        """
        pass

    @staticmethod
    def _futures_expiry(dt: datetime = datetime.now(), months: int = 1) -> str:
        """Returns a string of format YYYYMM corresponding to the leading
        contract month, from the inputted datetime object.

        Parameters
        ----------
        dt : datetime, optional
            The datetime object to convert to string. The default is datetime.now().
        months : int, optional
            The month offset from dt. The default is 1.

        Returns
        -------
        str
            A string corresponding to the expiry.
        """
        # TODO - if current date is past last trade date... error
        expiry_dt = dt + relativedelta(months=months)
        return expiry_dt.strftime("%Y") + expiry_dt.strftime("%m")

    def get_precision(self, instrument, *args, **kwargs):
        """Returns the precision of the instrument."""
        # NOTE: this has not been formally implemented yet.
        # TODO - implement
        return {"size": 5, "price": 5}
